/******************************************************************************
 * src/algorithms/networks.h
 *
 * Sorting networks for small fixed sizes
 *
 * Generated using http://pages.ripco.net/~jgamble/nw.html
 *
 ******************************************************************************
 * Copyright (C) 2014 Timo Bingmann <tb@panthema.net>
 *
 * This program is free software: you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the Free
 * Software Foundation, either version 3 of the License, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
 * more details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program.  If not, see <http://www.gnu.org/licenses/>.
 *****************************************************************************/

namespace networks {

template <typename ValueType>
static inline
void SWAP(ValueType& A, ValueType& B)
{
    if (A > B) { std::swap(A, B); }
}

template <typename ValueType>
static inline
void sort2(ValueType& A0, ValueType& A1)
{
    SWAP(A0, A1);
}

template <typename ValueType>
static inline
void sort3(ValueType& A0, ValueType& A1, ValueType& A2)
{
    SWAP(A1, A2);
    SWAP(A0, A2);
    SWAP(A0, A1);
}

template <typename ValueType>
static inline
void sort4(ValueType& A0, ValueType& A1, ValueType& A2, ValueType& A3)
{
    SWAP(A0, A1);
    SWAP(A2, A3);
    SWAP(A0, A2);
    SWAP(A1, A3);
    SWAP(A1, A2);
}

template <typename ValueType>
static inline
void sort5(ValueType& A0, ValueType& A1, ValueType& A2, ValueType& A3, ValueType& A4)
{
    SWAP(A0, A1);
    SWAP(A3, A4);
    SWAP(A2, A4);
    SWAP(A2, A3);
    SWAP(A0, A3);
    SWAP(A0, A2);
    SWAP(A1, A4);
    SWAP(A1, A3);
    SWAP(A1, A2);
}

template <typename ValueType>
static inline
void sort6(ValueType& A0, ValueType& A1, ValueType& A2, ValueType& A3, ValueType& A4, ValueType& A5)
{
    SWAP(A1, A2);
    SWAP(A0, A2);
    SWAP(A0, A1);
    SWAP(A4, A5);
    SWAP(A3, A5);
    SWAP(A3, A4);
    SWAP(A0, A3);
    SWAP(A1, A4);
    SWAP(A2, A5);
    SWAP(A2, A4);
    SWAP(A1, A3);
    SWAP(A2, A3);
}

template <typename ValueType>
static inline
void sort7(ValueType& A0, ValueType& A1, ValueType& A2, ValueType& A3, ValueType& A4, ValueType& A5, ValueType& A6)
{
    SWAP(A1, A2);
    SWAP(A0, A2);
    SWAP(A0, A1);
    SWAP(A3, A4);
    SWAP(A5, A6);
    SWAP(A3, A5);
    SWAP(A4, A6);
    SWAP(A4, A5);
    SWAP(A0, A4);
    SWAP(A0, A3);
    SWAP(A1, A5);
    SWAP(A2, A6);
    SWAP(A2, A5);
    SWAP(A1, A3);
    SWAP(A2, A4);
    SWAP(A2, A3);
}

template <typename ValueType>
static inline
void sort8(ValueType& A0, ValueType& A1, ValueType& A2, ValueType& A3, ValueType& A4, ValueType& A5, ValueType& A6, ValueType& A7)
{
    SWAP(A0, A1);
    SWAP(A2, A3);
    SWAP(A0, A2);
    SWAP(A1, A3);
    SWAP(A1, A2);
    SWAP(A4, A5);
    SWAP(A6, A7);
    SWAP(A4, A6);
    SWAP(A5, A7);
    SWAP(A5, A6);
    SWAP(A0, A4);
    SWAP(A1, A5);
    SWAP(A1, A4);
    SWAP(A2, A6);
    SWAP(A3, A7);
    SWAP(A3, A6);
    SWAP(A2, A4);
    SWAP(A3, A5);
    SWAP(A3, A4);
}

template <typename ValueType>
static inline
void sort9(ValueType& A0, ValueType& A1, ValueType& A2, ValueType& A3, ValueType& A4, ValueType& A5, ValueType& A6, ValueType& A7, ValueType& A8)
{
    SWAP(A0, A1);
    SWAP(A3, A4);
    SWAP(A6, A7);
    SWAP(A1, A2);
    SWAP(A4, A5);
    SWAP(A7, A8);
    SWAP(A0, A1);
    SWAP(A3, A4);
    SWAP(A6, A7);
    SWAP(A0, A3);
    SWAP(A3, A6);
    SWAP(A0, A3);
    SWAP(A1, A4);
    SWAP(A4, A7);
    SWAP(A1, A4);
    SWAP(A2, A5);
    SWAP(A5, A8);
    SWAP(A2, A5);
    SWAP(A1, A3);
    SWAP(A5, A7);
    SWAP(A2, A6);
    SWAP(A4, A6);
    SWAP(A2, A4);
    SWAP(A2, A3);
    SWAP(A5, A6);
}

template <typename ValueType>
static inline
void sort10(ValueType& A0, ValueType& A1, ValueType& A2, ValueType& A3, ValueType& A4, ValueType& A5, ValueType& A6, ValueType& A7, ValueType& A8, ValueType& A9)
{
    SWAP(A4, A9);
    SWAP(A3, A8);
    SWAP(A2, A7);
    SWAP(A1, A6);
    SWAP(A0, A5);
    SWAP(A1, A4);
    SWAP(A6, A9);
    SWAP(A0, A3);
    SWAP(A5, A8);
    SWAP(A0, A2);
    SWAP(A3, A6);
    SWAP(A7, A9);
    SWAP(A0, A1);
    SWAP(A2, A4);
    SWAP(A5, A7);
    SWAP(A8, A9);
    SWAP(A1, A2);
    SWAP(A4, A6);
    SWAP(A7, A8);
    SWAP(A3, A5);
    SWAP(A2, A5);
    SWAP(A6, A8);
    SWAP(A1, A3);
    SWAP(A4, A7);
    SWAP(A2, A3);
    SWAP(A6, A7);
    SWAP(A3, A4);
    SWAP(A5, A6);
    SWAP(A4, A5);
}

template <typename ValueType>
static inline
void sort11(ValueType& A0, ValueType& A1, ValueType& A2, ValueType& A3, ValueType& A4, ValueType& A5, ValueType& A6, ValueType& A7, ValueType& A8, ValueType& A9, ValueType& A10)
{
    SWAP(A0, A1);
    SWAP(A2, A3);
    SWAP(A4, A5);
    SWAP(A6, A7);
    SWAP(A8, A9);
    SWAP(A1, A3);
    SWAP(A5, A7);
    SWAP(A0, A2);
    SWAP(A4, A6);
    SWAP(A8, A10);
    SWAP(A1, A2);
    SWAP(A5, A6);
    SWAP(A9, A10);
    SWAP(A1, A5);
    SWAP(A6, A10);
    SWAP(A5, A9);
    SWAP(A2, A6);
    SWAP(A1, A5);
    SWAP(A6, A10);
    SWAP(A0, A4);
    SWAP(A3, A7);
    SWAP(A4, A8);
    SWAP(A0, A4);
    SWAP(A1, A4);
    SWAP(A7, A10);
    SWAP(A3, A8);
    SWAP(A2, A3);
    SWAP(A8, A9);
    SWAP(A2, A4);
    SWAP(A7, A9);
    SWAP(A3, A5);
    SWAP(A6, A8);
    SWAP(A3, A4);
    SWAP(A5, A6);
    SWAP(A7, A8);
}

template <typename ValueType>
static inline
void sort12(ValueType& A0, ValueType& A1, ValueType& A2, ValueType& A3, ValueType& A4, ValueType& A5, ValueType& A6, ValueType& A7, ValueType& A8, ValueType& A9, ValueType& A10, ValueType& A11)
{
    SWAP(A0, A1);
    SWAP(A2, A3);
    SWAP(A4, A5);
    SWAP(A6, A7);
    SWAP(A8, A9);
    SWAP(A10, A11);
    SWAP(A1, A3);
    SWAP(A5, A7);
    SWAP(A9, A11);
    SWAP(A0, A2);
    SWAP(A4, A6);
    SWAP(A8, A10);
    SWAP(A1, A2);
    SWAP(A5, A6);
    SWAP(A9, A10);
    SWAP(A1, A5);
    SWAP(A6, A10);
    SWAP(A5, A9);
    SWAP(A2, A6);
    SWAP(A1, A5);
    SWAP(A6, A10);
    SWAP(A0, A4);
    SWAP(A7, A11);
    SWAP(A3, A7);
    SWAP(A4, A8);
    SWAP(A0, A4);
    SWAP(A7, A11);
    SWAP(A1, A4);
    SWAP(A7, A10);
    SWAP(A3, A8);
    SWAP(A2, A3);
    SWAP(A8, A9);
    SWAP(A2, A4);
    SWAP(A7, A9);
    SWAP(A3, A5);
    SWAP(A6, A8);
    SWAP(A3, A4);
    SWAP(A5, A6);
    SWAP(A7, A8);
}

template <typename ValueType>
static inline
void sort13(ValueType& A0, ValueType& A1, ValueType& A2, ValueType& A3, ValueType& A4, ValueType& A5, ValueType& A6, ValueType& A7, ValueType& A8, ValueType& A9, ValueType& A10, ValueType& A11, ValueType& A12)
{
    SWAP(A1, A7);
    SWAP(A9, A11);
    SWAP(A3, A4);
    SWAP(A5, A8);
    SWAP(A0, A12);
    SWAP(A2, A6);
    SWAP(A0, A1);
    SWAP(A2, A3);
    SWAP(A4, A6);
    SWAP(A8, A11);
    SWAP(A7, A12);
    SWAP(A5, A9);
    SWAP(A0, A2);
    SWAP(A3, A7);
    SWAP(A10, A11);
    SWAP(A1, A4);
    SWAP(A6, A12);
    SWAP(A7, A8);
    SWAP(A11, A12);
    SWAP(A4, A9);
    SWAP(A6, A10);
    SWAP(A3, A4);
    SWAP(A5, A6);
    SWAP(A8, A9);
    SWAP(A10, A11);
    SWAP(A1, A7);
    SWAP(A2, A6);
    SWAP(A9, A11);
    SWAP(A1, A3);
    SWAP(A4, A7);
    SWAP(A8, A10);
    SWAP(A0, A5);
    SWAP(A2, A5);
    SWAP(A6, A8);
    SWAP(A9, A10);
    SWAP(A1, A2);
    SWAP(A3, A5);
    SWAP(A7, A8);
    SWAP(A4, A6);
    SWAP(A2, A3);
    SWAP(A4, A5);
    SWAP(A6, A7);
    SWAP(A8, A9);
    SWAP(A3, A4);
    SWAP(A5, A6);
}

template <typename ValueType>
static inline
void sort14(ValueType& A0, ValueType& A1, ValueType& A2, ValueType& A3, ValueType& A4, ValueType& A5, ValueType& A6, ValueType& A7, ValueType& A8, ValueType& A9, ValueType& A10, ValueType& A11, ValueType& A12, ValueType& A13)
{
    SWAP(A0, A1);
    SWAP(A2, A3);
    SWAP(A4, A5);
    SWAP(A6, A7);
    SWAP(A8, A9);
    SWAP(A10, A11);
    SWAP(A12, A13);
    SWAP(A0, A2);
    SWAP(A4, A6);
    SWAP(A8, A10);
    SWAP(A1, A3);
    SWAP(A5, A7);
    SWAP(A9, A11);
    SWAP(A0, A4);
    SWAP(A8, A12);
    SWAP(A1, A5);
    SWAP(A9, A13);
    SWAP(A2, A6);
    SWAP(A3, A7);
    SWAP(A0, A8);
    SWAP(A1, A9);
    SWAP(A2, A10);
    SWAP(A3, A11);
    SWAP(A4, A12);
    SWAP(A5, A13);
    SWAP(A5, A10);
    SWAP(A6, A9);
    SWAP(A3, A12);
    SWAP(A7, A11);
    SWAP(A1, A2);
    SWAP(A4, A8);
    SWAP(A1, A4);
    SWAP(A7, A13);
    SWAP(A2, A8);
    SWAP(A2, A4);
    SWAP(A5, A6);
    SWAP(A9, A10);
    SWAP(A11, A13);
    SWAP(A3, A8);
    SWAP(A7, A12);
    SWAP(A6, A8);
    SWAP(A10, A12);
    SWAP(A3, A5);
    SWAP(A7, A9);
    SWAP(A3, A4);
    SWAP(A5, A6);
    SWAP(A7, A8);
    SWAP(A9, A10);
    SWAP(A11, A12);
    SWAP(A6, A7);
    SWAP(A8, A9);
}

template <typename ValueType>
static inline
void sort15(ValueType& A0, ValueType& A1, ValueType& A2, ValueType& A3, ValueType& A4, ValueType& A5, ValueType& A6, ValueType& A7, ValueType& A8, ValueType& A9, ValueType& A10, ValueType& A11, ValueType& A12, ValueType& A13, ValueType& A14)
{
    SWAP(A0, A1);
    SWAP(A2, A3);
    SWAP(A4, A5);
    SWAP(A6, A7);
    SWAP(A8, A9);
    SWAP(A10, A11);
    SWAP(A12, A13);
    SWAP(A0, A2);
    SWAP(A4, A6);
    SWAP(A8, A10);
    SWAP(A12, A14);
    SWAP(A1, A3);
    SWAP(A5, A7);
    SWAP(A9, A11);
    SWAP(A0, A4);
    SWAP(A8, A12);
    SWAP(A1, A5);
    SWAP(A9, A13);
    SWAP(A2, A6);
    SWAP(A10, A14);
    SWAP(A3, A7);
    SWAP(A0, A8);
    SWAP(A1, A9);
    SWAP(A2, A10);
    SWAP(A3, A11);
    SWAP(A4, A12);
    SWAP(A5, A13);
    SWAP(A6, A14);
    SWAP(A5, A10);
    SWAP(A6, A9);
    SWAP(A3, A12);
    SWAP(A13, A14);
    SWAP(A7, A11);
    SWAP(A1, A2);
    SWAP(A4, A8);
    SWAP(A1, A4);
    SWAP(A7, A13);
    SWAP(A2, A8);
    SWAP(A11, A14);
    SWAP(A2, A4);
    SWAP(A5, A6);
    SWAP(A9, A10);
    SWAP(A11, A13);
    SWAP(A3, A8);
    SWAP(A7, A12);
    SWAP(A6, A8);
    SWAP(A10, A12);
    SWAP(A3, A5);
    SWAP(A7, A9);
    SWAP(A3, A4);
    SWAP(A5, A6);
    SWAP(A7, A8);
    SWAP(A9, A10);
    SWAP(A11, A12);
    SWAP(A6, A7);
    SWAP(A8, A9);
}

template <typename ValueType>
static inline
void sort16(ValueType& A0, ValueType& A1, ValueType& A2, ValueType& A3, ValueType& A4, ValueType& A5, ValueType& A6, ValueType& A7, ValueType& A8, ValueType& A9, ValueType& A10, ValueType& A11, ValueType& A12, ValueType& A13, ValueType& A14, ValueType& A15)
{
    SWAP(A0, A1);
    SWAP(A2, A3);
    SWAP(A4, A5);
    SWAP(A6, A7);
    SWAP(A8, A9);
    SWAP(A10, A11);
    SWAP(A12, A13);
    SWAP(A14, A15);
    SWAP(A0, A2);
    SWAP(A4, A6);
    SWAP(A8, A10);
    SWAP(A12, A14);
    SWAP(A1, A3);
    SWAP(A5, A7);
    SWAP(A9, A11);
    SWAP(A13, A15);
    SWAP(A0, A4);
    SWAP(A8, A12);
    SWAP(A1, A5);
    SWAP(A9, A13);
    SWAP(A2, A6);
    SWAP(A10, A14);
    SWAP(A3, A7);
    SWAP(A11, A15);
    SWAP(A0, A8);
    SWAP(A1, A9);
    SWAP(A2, A10);
    SWAP(A3, A11);
    SWAP(A4, A12);
    SWAP(A5, A13);
    SWAP(A6, A14);
    SWAP(A7, A15);
    SWAP(A5, A10);
    SWAP(A6, A9);
    SWAP(A3, A12);
    SWAP(A13, A14);
    SWAP(A7, A11);
    SWAP(A1, A2);
    SWAP(A4, A8);
    SWAP(A1, A4);
    SWAP(A7, A13);
    SWAP(A2, A8);
    SWAP(A11, A14);
    SWAP(A2, A4);
    SWAP(A5, A6);
    SWAP(A9, A10);
    SWAP(A11, A13);
    SWAP(A3, A8);
    SWAP(A7, A12);
    SWAP(A6, A8);
    SWAP(A10, A12);
    SWAP(A3, A5);
    SWAP(A7, A9);
    SWAP(A3, A4);
    SWAP(A5, A6);
    SWAP(A7, A8);
    SWAP(A9, A10);
    SWAP(A11, A12);
    SWAP(A6, A7);
    SWAP(A8, A9);
}

template <typename ValueType>
static inline
void sort17(ValueType& A0, ValueType& A1, ValueType& A2, ValueType& A3, ValueType& A4, ValueType& A5, ValueType& A6, ValueType& A7, ValueType& A8, ValueType& A9, ValueType& A10, ValueType& A11, ValueType& A12, ValueType& A13, ValueType& A14, ValueType& A15, ValueType& A16)
{
    SWAP(A0, A1);
    SWAP(A2, A3);
    SWAP(A0, A2);
    SWAP(A1, A3);
    SWAP(A1, A2);
    SWAP(A4, A5);
    SWAP(A6, A7);
    SWAP(A4, A6);
    SWAP(A5, A7);
    SWAP(A5, A6);
    SWAP(A0, A4);
    SWAP(A1, A5);
    SWAP(A1, A4);
    SWAP(A2, A6);
    SWAP(A3, A7);
    SWAP(A3, A6);
    SWAP(A2, A4);
    SWAP(A3, A5);
    SWAP(A3, A4);
    SWAP(A8, A9);
    SWAP(A10, A11);
    SWAP(A8, A10);
    SWAP(A9, A11);
    SWAP(A9, A10);
    SWAP(A12, A13);
    SWAP(A15, A16);
    SWAP(A14, A16);
    SWAP(A14, A15);
    SWAP(A12, A15);
    SWAP(A12, A14);
    SWAP(A13, A16);
    SWAP(A13, A15);
    SWAP(A13, A14);
    SWAP(A8, A13);
    SWAP(A8, A12);
    SWAP(A9, A14);
    SWAP(A9, A13);
    SWAP(A9, A12);
    SWAP(A10, A15);
    SWAP(A11, A16);
    SWAP(A11, A15);
    SWAP(A10, A13);
    SWAP(A10, A12);
    SWAP(A11, A14);
    SWAP(A11, A13);
    SWAP(A11, A12);
    SWAP(A0, A9);
    SWAP(A0, A8);
    SWAP(A1, A10);
    SWAP(A1, A9);
    SWAP(A1, A8);
    SWAP(A2, A11);
    SWAP(A3, A12);
    SWAP(A3, A11);
    SWAP(A2, A9);
    SWAP(A2, A8);
    SWAP(A3, A10);
    SWAP(A3, A9);
    SWAP(A3, A8);
    SWAP(A4, A13);
    SWAP(A5, A14);
    SWAP(A5, A13);
    SWAP(A6, A15);
    SWAP(A7, A16);
    SWAP(A7, A15);
    SWAP(A6, A13);
    SWAP(A7, A14);
    SWAP(A7, A13);
    SWAP(A4, A9);
    SWAP(A4, A8);
    SWAP(A5, A10);
    SWAP(A5, A9);
    SWAP(A5, A8);
    SWAP(A6, A11);
    SWAP(A7, A12);
    SWAP(A7, A11);
    SWAP(A6, A9);
    SWAP(A6, A8);
    SWAP(A7, A10);
    SWAP(A7, A9);
    SWAP(A7, A8);
}

template <typename ValueType>
static inline
void sort18(ValueType& A0, ValueType& A1, ValueType& A2, ValueType& A3, ValueType& A4, ValueType& A5, ValueType& A6, ValueType& A7, ValueType& A8, ValueType& A9, ValueType& A10, ValueType& A11, ValueType& A12, ValueType& A13, ValueType& A14, ValueType& A15, ValueType& A16, ValueType& A17)
{
    SWAP(A0, A1);
    SWAP(A2, A3);
    SWAP(A0, A2);
    SWAP(A1, A3);
    SWAP(A1, A2);
    SWAP(A4, A5);
    SWAP(A7, A8);
    SWAP(A6, A8);
    SWAP(A6, A7);
    SWAP(A4, A7);
    SWAP(A4, A6);
    SWAP(A5, A8);
    SWAP(A5, A7);
    SWAP(A5, A6);
    SWAP(A0, A5);
    SWAP(A0, A4);
    SWAP(A1, A6);
    SWAP(A1, A5);
    SWAP(A1, A4);
    SWAP(A2, A7);
    SWAP(A3, A8);
    SWAP(A3, A7);
    SWAP(A2, A5);
    SWAP(A2, A4);
    SWAP(A3, A6);
    SWAP(A3, A5);
    SWAP(A3, A4);
    SWAP(A9, A10);
    SWAP(A11, A12);
    SWAP(A9, A11);
    SWAP(A10, A12);
    SWAP(A10, A11);
    SWAP(A13, A14);
    SWAP(A16, A17);
    SWAP(A15, A17);
    SWAP(A15, A16);
    SWAP(A13, A16);
    SWAP(A13, A15);
    SWAP(A14, A17);
    SWAP(A14, A16);
    SWAP(A14, A15);
    SWAP(A9, A14);
    SWAP(A9, A13);
    SWAP(A10, A15);
    SWAP(A10, A14);
    SWAP(A10, A13);
    SWAP(A11, A16);
    SWAP(A12, A17);
    SWAP(A12, A16);
    SWAP(A11, A14);
    SWAP(A11, A13);
    SWAP(A12, A15);
    SWAP(A12, A14);
    SWAP(A12, A13);
    SWAP(A0, A9);
    SWAP(A1, A10);
    SWAP(A1, A9);
    SWAP(A2, A11);
    SWAP(A3, A12);
    SWAP(A3, A11);
    SWAP(A2, A9);
    SWAP(A3, A10);
    SWAP(A3, A9);
    SWAP(A4, A13);
    SWAP(A5, A14);
    SWAP(A5, A13);
    SWAP(A6, A15);
    SWAP(A7, A16);
    SWAP(A8, A17);
    SWAP(A8, A16);
    SWAP(A7, A15);
    SWAP(A8, A15);
    SWAP(A6, A13);
    SWAP(A7, A14);
    SWAP(A8, A14);
    SWAP(A7, A13);
    SWAP(A8, A13);
    SWAP(A4, A9);
    SWAP(A5, A10);
    SWAP(A5, A9);
    SWAP(A6, A11);
    SWAP(A7, A12);
    SWAP(A8, A12);
    SWAP(A7, A11);
    SWAP(A8, A11);
    SWAP(A6, A9);
    SWAP(A7, A10);
    SWAP(A8, A10);
    SWAP(A7, A9);
    SWAP(A8, A9);
}

template <typename ValueType>
static inline
void sort19(ValueType& A0, ValueType& A1, ValueType& A2, ValueType& A3, ValueType& A4, ValueType& A5, ValueType& A6, ValueType& A7, ValueType& A8, ValueType& A9, ValueType& A10, ValueType& A11, ValueType& A12, ValueType& A13, ValueType& A14, ValueType& A15, ValueType& A16, ValueType& A17, ValueType& A18)
{
    SWAP(A0, A1);
    SWAP(A2, A3);
    SWAP(A0, A2);
    SWAP(A1, A3);
    SWAP(A1, A2);
    SWAP(A4, A5);
    SWAP(A7, A8);
    SWAP(A6, A8);
    SWAP(A6, A7);
    SWAP(A4, A7);
    SWAP(A4, A6);
    SWAP(A5, A8);
    SWAP(A5, A7);
    SWAP(A5, A6);
    SWAP(A0, A5);
    SWAP(A0, A4);
    SWAP(A1, A6);
    SWAP(A1, A5);
    SWAP(A1, A4);
    SWAP(A2, A7);
    SWAP(A3, A8);
    SWAP(A3, A7);
    SWAP(A2, A5);
    SWAP(A2, A4);
    SWAP(A3, A6);
    SWAP(A3, A5);
    SWAP(A3, A4);
    SWAP(A9, A10);
    SWAP(A12, A13);
    SWAP(A11, A13);
    SWAP(A11, A12);
    SWAP(A9, A12);
    SWAP(A9, A11);
    SWAP(A10, A13);
    SWAP(A10, A12);
    SWAP(A10, A11);
    SWAP(A14, A15);
    SWAP(A17, A18);
    SWAP(A16, A18);
    SWAP(A16, A17);
    SWAP(A14, A17);
    SWAP(A14, A16);
    SWAP(A15, A18);
    SWAP(A15, A17);
    SWAP(A15, A16);
    SWAP(A9, A14);
    SWAP(A10, A15);
    SWAP(A10, A14);
    SWAP(A11, A16);
    SWAP(A12, A17);
    SWAP(A13, A18);
    SWAP(A13, A17);
    SWAP(A12, A16);
    SWAP(A13, A16);
    SWAP(A11, A14);
    SWAP(A12, A15);
    SWAP(A13, A15);
    SWAP(A12, A14);
    SWAP(A13, A14);
    SWAP(A0, A10);
    SWAP(A0, A9);
    SWAP(A1, A11);
    SWAP(A1, A10);
    SWAP(A1, A9);
    SWAP(A2, A12);
    SWAP(A3, A13);
    SWAP(A3, A12);
    SWAP(A2, A10);
    SWAP(A2, A9);
    SWAP(A3, A11);
    SWAP(A3, A10);
    SWAP(A3, A9);
    SWAP(A4, A14);
    SWAP(A5, A15);
    SWAP(A5, A14);
    SWAP(A6, A16);
    SWAP(A7, A17);
    SWAP(A8, A18);
    SWAP(A8, A17);
    SWAP(A7, A16);
    SWAP(A8, A16);
    SWAP(A6, A14);
    SWAP(A7, A15);
    SWAP(A8, A15);
    SWAP(A7, A14);
    SWAP(A8, A14);
    SWAP(A4, A9);
    SWAP(A5, A10);
    SWAP(A5, A9);
    SWAP(A6, A11);
    SWAP(A7, A12);
    SWAP(A8, A13);
    SWAP(A8, A12);
    SWAP(A7, A11);
    SWAP(A8, A11);
    SWAP(A6, A9);
    SWAP(A7, A10);
    SWAP(A8, A10);
    SWAP(A7, A9);
    SWAP(A8, A9);
}

template <typename ValueType>
static inline
void sort20(ValueType& A0, ValueType& A1, ValueType& A2, ValueType& A3, ValueType& A4, ValueType& A5, ValueType& A6, ValueType& A7, ValueType& A8, ValueType& A9, ValueType& A10, ValueType& A11, ValueType& A12, ValueType& A13, ValueType& A14, ValueType& A15, ValueType& A16, ValueType& A17, ValueType& A18, ValueType& A19)
{
    SWAP(A0, A1);
    SWAP(A3, A4);
    SWAP(A2, A4);
    SWAP(A2, A3);
    SWAP(A0, A3);
    SWAP(A0, A2);
    SWAP(A1, A4);
    SWAP(A1, A3);
    SWAP(A1, A2);
    SWAP(A5, A6);
    SWAP(A8, A9);
    SWAP(A7, A9);
    SWAP(A7, A8);
    SWAP(A5, A8);
    SWAP(A5, A7);
    SWAP(A6, A9);
    SWAP(A6, A8);
    SWAP(A6, A7);
    SWAP(A0, A5);
    SWAP(A1, A6);
    SWAP(A1, A5);
    SWAP(A2, A7);
    SWAP(A3, A8);
    SWAP(A4, A9);
    SWAP(A4, A8);
    SWAP(A3, A7);
    SWAP(A4, A7);
    SWAP(A2, A5);
    SWAP(A3, A6);
    SWAP(A4, A6);
    SWAP(A3, A5);
    SWAP(A4, A5);
    SWAP(A10, A11);
    SWAP(A13, A14);
    SWAP(A12, A14);
    SWAP(A12, A13);
    SWAP(A10, A13);
    SWAP(A10, A12);
    SWAP(A11, A14);
    SWAP(A11, A13);
    SWAP(A11, A12);
    SWAP(A15, A16);
    SWAP(A18, A19);
    SWAP(A17, A19);
    SWAP(A17, A18);
    SWAP(A15, A18);
    SWAP(A15, A17);
    SWAP(A16, A19);
    SWAP(A16, A18);
    SWAP(A16, A17);
    SWAP(A10, A15);
    SWAP(A11, A16);
    SWAP(A11, A15);
    SWAP(A12, A17);
    SWAP(A13, A18);
    SWAP(A14, A19);
    SWAP(A14, A18);
    SWAP(A13, A17);
    SWAP(A14, A17);
    SWAP(A12, A15);
    SWAP(A13, A16);
    SWAP(A14, A16);
    SWAP(A13, A15);
    SWAP(A14, A15);
    SWAP(A0, A10);
    SWAP(A1, A11);
    SWAP(A1, A10);
    SWAP(A2, A12);
    SWAP(A3, A13);
    SWAP(A4, A14);
    SWAP(A4, A13);
    SWAP(A3, A12);
    SWAP(A4, A12);
    SWAP(A2, A10);
    SWAP(A3, A11);
    SWAP(A4, A11);
    SWAP(A3, A10);
    SWAP(A4, A10);
    SWAP(A5, A15);
    SWAP(A6, A16);
    SWAP(A6, A15);
    SWAP(A7, A17);
    SWAP(A8, A18);
    SWAP(A9, A19);
    SWAP(A9, A18);
    SWAP(A8, A17);
    SWAP(A9, A17);
    SWAP(A7, A15);
    SWAP(A8, A16);
    SWAP(A9, A16);
    SWAP(A8, A15);
    SWAP(A9, A15);
    SWAP(A5, A10);
    SWAP(A6, A11);
    SWAP(A6, A10);
    SWAP(A7, A12);
    SWAP(A8, A13);
    SWAP(A9, A14);
    SWAP(A9, A13);
    SWAP(A8, A12);
    SWAP(A9, A12);
    SWAP(A7, A10);
    SWAP(A8, A11);
    SWAP(A9, A11);
    SWAP(A8, A10);
    SWAP(A9, A10);
}

template <typename ValueType>
static inline
void sort21(ValueType& A0, ValueType& A1, ValueType& A2, ValueType& A3, ValueType& A4, ValueType& A5, ValueType& A6, ValueType& A7, ValueType& A8, ValueType& A9, ValueType& A10, ValueType& A11, ValueType& A12, ValueType& A13, ValueType& A14, ValueType& A15, ValueType& A16, ValueType& A17, ValueType& A18, ValueType& A19, ValueType& A20)
{
    SWAP(A0, A1);
    SWAP(A3, A4);
    SWAP(A2, A4);
    SWAP(A2, A3);
    SWAP(A0, A3);
    SWAP(A0, A2);
    SWAP(A1, A4);
    SWAP(A1, A3);
    SWAP(A1, A2);
    SWAP(A5, A6);
    SWAP(A8, A9);
    SWAP(A7, A9);
    SWAP(A7, A8);
    SWAP(A5, A8);
    SWAP(A5, A7);
    SWAP(A6, A9);
    SWAP(A6, A8);
    SWAP(A6, A7);
    SWAP(A0, A5);
    SWAP(A1, A6);
    SWAP(A1, A5);
    SWAP(A2, A7);
    SWAP(A3, A8);
    SWAP(A4, A9);
    SWAP(A4, A8);
    SWAP(A3, A7);
    SWAP(A4, A7);
    SWAP(A2, A5);
    SWAP(A3, A6);
    SWAP(A4, A6);
    SWAP(A3, A5);
    SWAP(A4, A5);
    SWAP(A10, A11);
    SWAP(A13, A14);
    SWAP(A12, A14);
    SWAP(A12, A13);
    SWAP(A10, A13);
    SWAP(A10, A12);
    SWAP(A11, A14);
    SWAP(A11, A13);
    SWAP(A11, A12);
    SWAP(A16, A17);
    SWAP(A15, A17);
    SWAP(A15, A16);
    SWAP(A19, A20);
    SWAP(A18, A20);
    SWAP(A18, A19);
    SWAP(A15, A18);
    SWAP(A16, A19);
    SWAP(A17, A20);
    SWAP(A17, A19);
    SWAP(A16, A18);
    SWAP(A17, A18);
    SWAP(A10, A16);
    SWAP(A10, A15);
    SWAP(A11, A17);
    SWAP(A11, A16);
    SWAP(A11, A15);
    SWAP(A12, A18);
    SWAP(A13, A19);
    SWAP(A14, A20);
    SWAP(A14, A19);
    SWAP(A13, A18);
    SWAP(A14, A18);
    SWAP(A12, A15);
    SWAP(A13, A16);
    SWAP(A14, A17);
    SWAP(A14, A16);
    SWAP(A13, A15);
    SWAP(A14, A15);
    SWAP(A0, A11);
    SWAP(A0, A10);
    SWAP(A1, A12);
    SWAP(A1, A11);
    SWAP(A1, A10);
    SWAP(A2, A13);
    SWAP(A3, A14);
    SWAP(A4, A15);
    SWAP(A4, A14);
    SWAP(A3, A13);
    SWAP(A4, A13);
    SWAP(A2, A10);
    SWAP(A3, A11);
    SWAP(A4, A12);
    SWAP(A4, A11);
    SWAP(A3, A10);
    SWAP(A4, A10);
    SWAP(A5, A16);
    SWAP(A6, A17);
    SWAP(A6, A16);
    SWAP(A7, A18);
    SWAP(A8, A19);
    SWAP(A9, A20);
    SWAP(A9, A19);
    SWAP(A8, A18);
    SWAP(A9, A18);
    SWAP(A7, A16);
    SWAP(A8, A17);
    SWAP(A9, A17);
    SWAP(A8, A16);
    SWAP(A9, A16);
    SWAP(A5, A11);
    SWAP(A5, A10);
    SWAP(A6, A12);
    SWAP(A6, A11);
    SWAP(A6, A10);
    SWAP(A7, A13);
    SWAP(A8, A14);
    SWAP(A9, A15);
    SWAP(A9, A14);
    SWAP(A8, A13);
    SWAP(A9, A13);
    SWAP(A7, A10);
    SWAP(A8, A11);
    SWAP(A9, A12);
    SWAP(A9, A11);
    SWAP(A8, A10);
    SWAP(A9, A10);
}

template <typename ValueType>
static inline
void sort22(ValueType& A0, ValueType& A1, ValueType& A2, ValueType& A3, ValueType& A4, ValueType& A5, ValueType& A6, ValueType& A7, ValueType& A8, ValueType& A9, ValueType& A10, ValueType& A11, ValueType& A12, ValueType& A13, ValueType& A14, ValueType& A15, ValueType& A16, ValueType& A17, ValueType& A18, ValueType& A19, ValueType& A20, ValueType& A21)
{
    SWAP(A0, A1);
    SWAP(A3, A4);
    SWAP(A2, A4);
    SWAP(A2, A3);
    SWAP(A0, A3);
    SWAP(A0, A2);
    SWAP(A1, A4);
    SWAP(A1, A3);
    SWAP(A1, A2);
    SWAP(A6, A7);
    SWAP(A5, A7);
    SWAP(A5, A6);
    SWAP(A9, A10);
    SWAP(A8, A10);
    SWAP(A8, A9);
    SWAP(A5, A8);
    SWAP(A6, A9);
    SWAP(A7, A10);
    SWAP(A7, A9);
    SWAP(A6, A8);
    SWAP(A7, A8);
    SWAP(A0, A6);
    SWAP(A0, A5);
    SWAP(A1, A7);
    SWAP(A1, A6);
    SWAP(A1, A5);
    SWAP(A2, A8);
    SWAP(A3, A9);
    SWAP(A4, A10);
    SWAP(A4, A9);
    SWAP(A3, A8);
    SWAP(A4, A8);
    SWAP(A2, A5);
    SWAP(A3, A6);
    SWAP(A4, A7);
    SWAP(A4, A6);
    SWAP(A3, A5);
    SWAP(A4, A5);
    SWAP(A11, A12);
    SWAP(A14, A15);
    SWAP(A13, A15);
    SWAP(A13, A14);
    SWAP(A11, A14);
    SWAP(A11, A13);
    SWAP(A12, A15);
    SWAP(A12, A14);
    SWAP(A12, A13);
    SWAP(A17, A18);
    SWAP(A16, A18);
    SWAP(A16, A17);
    SWAP(A20, A21);
    SWAP(A19, A21);
    SWAP(A19, A20);
    SWAP(A16, A19);
    SWAP(A17, A20);
    SWAP(A18, A21);
    SWAP(A18, A20);
    SWAP(A17, A19);
    SWAP(A18, A19);
    SWAP(A11, A17);
    SWAP(A11, A16);
    SWAP(A12, A18);
    SWAP(A12, A17);
    SWAP(A12, A16);
    SWAP(A13, A19);
    SWAP(A14, A20);
    SWAP(A15, A21);
    SWAP(A15, A20);
    SWAP(A14, A19);
    SWAP(A15, A19);
    SWAP(A13, A16);
    SWAP(A14, A17);
    SWAP(A15, A18);
    SWAP(A15, A17);
    SWAP(A14, A16);
    SWAP(A15, A16);
    SWAP(A0, A11);
    SWAP(A1, A12);
    SWAP(A1, A11);
    SWAP(A2, A13);
    SWAP(A3, A14);
    SWAP(A4, A15);
    SWAP(A4, A14);
    SWAP(A3, A13);
    SWAP(A4, A13);
    SWAP(A2, A11);
    SWAP(A3, A12);
    SWAP(A4, A12);
    SWAP(A3, A11);
    SWAP(A4, A11);
    SWAP(A5, A16);
    SWAP(A6, A17);
    SWAP(A7, A18);
    SWAP(A7, A17);
    SWAP(A6, A16);
    SWAP(A7, A16);
    SWAP(A8, A19);
    SWAP(A9, A20);
    SWAP(A10, A21);
    SWAP(A10, A20);
    SWAP(A9, A19);
    SWAP(A10, A19);
    SWAP(A8, A16);
    SWAP(A9, A17);
    SWAP(A10, A18);
    SWAP(A10, A17);
    SWAP(A9, A16);
    SWAP(A10, A16);
    SWAP(A5, A11);
    SWAP(A6, A12);
    SWAP(A7, A13);
    SWAP(A7, A12);
    SWAP(A6, A11);
    SWAP(A7, A11);
    SWAP(A8, A14);
    SWAP(A9, A15);
    SWAP(A10, A15);
    SWAP(A9, A14);
    SWAP(A10, A14);
    SWAP(A8, A11);
    SWAP(A9, A12);
    SWAP(A10, A13);
    SWAP(A10, A12);
    SWAP(A9, A11);
    SWAP(A10, A11);
}

template <typename ValueType>
static inline
void sort23(ValueType& A0, ValueType& A1, ValueType& A2, ValueType& A3, ValueType& A4, ValueType& A5, ValueType& A6, ValueType& A7, ValueType& A8, ValueType& A9, ValueType& A10, ValueType& A11, ValueType& A12, ValueType& A13, ValueType& A14, ValueType& A15, ValueType& A16, ValueType& A17, ValueType& A18, ValueType& A19, ValueType& A20, ValueType& A21, ValueType& A22)
{
    SWAP(A0, A1);
    SWAP(A3, A4);
    SWAP(A2, A4);
    SWAP(A2, A3);
    SWAP(A0, A3);
    SWAP(A0, A2);
    SWAP(A1, A4);
    SWAP(A1, A3);
    SWAP(A1, A2);
    SWAP(A6, A7);
    SWAP(A5, A7);
    SWAP(A5, A6);
    SWAP(A9, A10);
    SWAP(A8, A10);
    SWAP(A8, A9);
    SWAP(A5, A8);
    SWAP(A6, A9);
    SWAP(A7, A10);
    SWAP(A7, A9);
    SWAP(A6, A8);
    SWAP(A7, A8);
    SWAP(A0, A6);
    SWAP(A0, A5);
    SWAP(A1, A7);
    SWAP(A1, A6);
    SWAP(A1, A5);
    SWAP(A2, A8);
    SWAP(A3, A9);
    SWAP(A4, A10);
    SWAP(A4, A9);
    SWAP(A3, A8);
    SWAP(A4, A8);
    SWAP(A2, A5);
    SWAP(A3, A6);
    SWAP(A4, A7);
    SWAP(A4, A6);
    SWAP(A3, A5);
    SWAP(A4, A5);
    SWAP(A12, A13);
    SWAP(A11, A13);
    SWAP(A11, A12);
    SWAP(A15, A16);
    SWAP(A14, A16);
    SWAP(A14, A15);
    SWAP(A11, A14);
    SWAP(A12, A15);
    SWAP(A13, A16);
    SWAP(A13, A15);
    SWAP(A12, A14);
    SWAP(A13, A14);
    SWAP(A18, A19);
    SWAP(A17, A19);
    SWAP(A17, A18);
    SWAP(A21, A22);
    SWAP(A20, A22);
    SWAP(A20, A21);
    SWAP(A17, A20);
    SWAP(A18, A21);
    SWAP(A19, A22);
    SWAP(A19, A21);
    SWAP(A18, A20);
    SWAP(A19, A20);
    SWAP(A11, A17);
    SWAP(A12, A18);
    SWAP(A13, A19);
    SWAP(A13, A18);
    SWAP(A12, A17);
    SWAP(A13, A17);
    SWAP(A14, A20);
    SWAP(A15, A21);
    SWAP(A16, A22);
    SWAP(A16, A21);
    SWAP(A15, A20);
    SWAP(A16, A20);
    SWAP(A14, A17);
    SWAP(A15, A18);
    SWAP(A16, A19);
    SWAP(A16, A18);
    SWAP(A15, A17);
    SWAP(A16, A17);
    SWAP(A0, A12);
    SWAP(A0, A11);
    SWAP(A1, A13);
    SWAP(A1, A12);
    SWAP(A1, A11);
    SWAP(A2, A14);
    SWAP(A3, A15);
    SWAP(A4, A16);
    SWAP(A4, A15);
    SWAP(A3, A14);
    SWAP(A4, A14);
    SWAP(A2, A11);
    SWAP(A3, A12);
    SWAP(A4, A13);
    SWAP(A4, A12);
    SWAP(A3, A11);
    SWAP(A4, A11);
    SWAP(A5, A17);
    SWAP(A6, A18);
    SWAP(A7, A19);
    SWAP(A7, A18);
    SWAP(A6, A17);
    SWAP(A7, A17);
    SWAP(A8, A20);
    SWAP(A9, A21);
    SWAP(A10, A22);
    SWAP(A10, A21);
    SWAP(A9, A20);
    SWAP(A10, A20);
    SWAP(A8, A17);
    SWAP(A9, A18);
    SWAP(A10, A19);
    SWAP(A10, A18);
    SWAP(A9, A17);
    SWAP(A10, A17);
    SWAP(A5, A11);
    SWAP(A6, A12);
    SWAP(A7, A13);
    SWAP(A7, A12);
    SWAP(A6, A11);
    SWAP(A7, A11);
    SWAP(A8, A14);
    SWAP(A9, A15);
    SWAP(A10, A16);
    SWAP(A10, A15);
    SWAP(A9, A14);
    SWAP(A10, A14);
    SWAP(A8, A11);
    SWAP(A9, A12);
    SWAP(A10, A13);
    SWAP(A10, A12);
    SWAP(A9, A11);
    SWAP(A10, A11);
}

template <typename ValueType>
static inline
void sort24(ValueType& A0, ValueType& A1, ValueType& A2, ValueType& A3, ValueType& A4, ValueType& A5, ValueType& A6, ValueType& A7, ValueType& A8, ValueType& A9, ValueType& A10, ValueType& A11, ValueType& A12, ValueType& A13, ValueType& A14, ValueType& A15, ValueType& A16, ValueType& A17, ValueType& A18, ValueType& A19, ValueType& A20, ValueType& A21, ValueType& A22, ValueType& A23)
{
    SWAP(A1, A2);
    SWAP(A0, A2);
    SWAP(A0, A1);
    SWAP(A4, A5);
    SWAP(A3, A5);
    SWAP(A3, A4);
    SWAP(A0, A3);
    SWAP(A1, A4);
    SWAP(A2, A5);
    SWAP(A2, A4);
    SWAP(A1, A3);
    SWAP(A2, A3);
    SWAP(A7, A8);
    SWAP(A6, A8);
    SWAP(A6, A7);
    SWAP(A10, A11);
    SWAP(A9, A11);
    SWAP(A9, A10);
    SWAP(A6, A9);
    SWAP(A7, A10);
    SWAP(A8, A11);
    SWAP(A8, A10);
    SWAP(A7, A9);
    SWAP(A8, A9);
    SWAP(A0, A6);
    SWAP(A1, A7);
    SWAP(A2, A8);
    SWAP(A2, A7);
    SWAP(A1, A6);
    SWAP(A2, A6);
    SWAP(A3, A9);
    SWAP(A4, A10);
    SWAP(A5, A11);
    SWAP(A5, A10);
    SWAP(A4, A9);
    SWAP(A5, A9);
    SWAP(A3, A6);
    SWAP(A4, A7);
    SWAP(A5, A8);
    SWAP(A5, A7);
    SWAP(A4, A6);
    SWAP(A5, A6);
    SWAP(A13, A14);
    SWAP(A12, A14);
    SWAP(A12, A13);
    SWAP(A16, A17);
    SWAP(A15, A17);
    SWAP(A15, A16);
    SWAP(A12, A15);
    SWAP(A13, A16);
    SWAP(A14, A17);
    SWAP(A14, A16);
    SWAP(A13, A15);
    SWAP(A14, A15);
    SWAP(A19, A20);
    SWAP(A18, A20);
    SWAP(A18, A19);
    SWAP(A22, A23);
    SWAP(A21, A23);
    SWAP(A21, A22);
    SWAP(A18, A21);
    SWAP(A19, A22);
    SWAP(A20, A23);
    SWAP(A20, A22);
    SWAP(A19, A21);
    SWAP(A20, A21);
    SWAP(A12, A18);
    SWAP(A13, A19);
    SWAP(A14, A20);
    SWAP(A14, A19);
    SWAP(A13, A18);
    SWAP(A14, A18);
    SWAP(A15, A21);
    SWAP(A16, A22);
    SWAP(A17, A23);
    SWAP(A17, A22);
    SWAP(A16, A21);
    SWAP(A17, A21);
    SWAP(A15, A18);
    SWAP(A16, A19);
    SWAP(A17, A20);
    SWAP(A17, A19);
    SWAP(A16, A18);
    SWAP(A17, A18);
    SWAP(A0, A12);
    SWAP(A1, A13);
    SWAP(A2, A14);
    SWAP(A2, A13);
    SWAP(A1, A12);
    SWAP(A2, A12);
    SWAP(A3, A15);
    SWAP(A4, A16);
    SWAP(A5, A17);
    SWAP(A5, A16);
    SWAP(A4, A15);
    SWAP(A5, A15);
    SWAP(A3, A12);
    SWAP(A4, A13);
    SWAP(A5, A14);
    SWAP(A5, A13);
    SWAP(A4, A12);
    SWAP(A5, A12);
    SWAP(A6, A18);
    SWAP(A7, A19);
    SWAP(A8, A20);
    SWAP(A8, A19);
    SWAP(A7, A18);
    SWAP(A8, A18);
    SWAP(A9, A21);
    SWAP(A10, A22);
    SWAP(A11, A23);
    SWAP(A11, A22);
    SWAP(A10, A21);
    SWAP(A11, A21);
    SWAP(A9, A18);
    SWAP(A10, A19);
    SWAP(A11, A20);
    SWAP(A11, A19);
    SWAP(A10, A18);
    SWAP(A11, A18);
    SWAP(A6, A12);
    SWAP(A7, A13);
    SWAP(A8, A14);
    SWAP(A8, A13);
    SWAP(A7, A12);
    SWAP(A8, A12);
    SWAP(A9, A15);
    SWAP(A10, A16);
    SWAP(A11, A17);
    SWAP(A11, A16);
    SWAP(A10, A15);
    SWAP(A11, A15);
    SWAP(A9, A12);
    SWAP(A10, A13);
    SWAP(A11, A14);
    SWAP(A11, A13);
    SWAP(A10, A12);
    SWAP(A11, A12);
}

template <typename ValueType>
static inline
void sort25(ValueType& A0, ValueType& A1, ValueType& A2, ValueType& A3, ValueType& A4, ValueType& A5, ValueType& A6, ValueType& A7, ValueType& A8, ValueType& A9, ValueType& A10, ValueType& A11, ValueType& A12, ValueType& A13, ValueType& A14, ValueType& A15, ValueType& A16, ValueType& A17, ValueType& A18, ValueType& A19, ValueType& A20, ValueType& A21, ValueType& A22, ValueType& A23, ValueType& A24)
{
    SWAP(A1, A2);
    SWAP(A0, A2);
    SWAP(A0, A1);
    SWAP(A4, A5);
    SWAP(A3, A5);
    SWAP(A3, A4);
    SWAP(A0, A3);
    SWAP(A1, A4);
    SWAP(A2, A5);
    SWAP(A2, A4);
    SWAP(A1, A3);
    SWAP(A2, A3);
    SWAP(A7, A8);
    SWAP(A6, A8);
    SWAP(A6, A7);
    SWAP(A10, A11);
    SWAP(A9, A11);
    SWAP(A9, A10);
    SWAP(A6, A9);
    SWAP(A7, A10);
    SWAP(A8, A11);
    SWAP(A8, A10);
    SWAP(A7, A9);
    SWAP(A8, A9);
    SWAP(A0, A6);
    SWAP(A1, A7);
    SWAP(A2, A8);
    SWAP(A2, A7);
    SWAP(A1, A6);
    SWAP(A2, A6);
    SWAP(A3, A9);
    SWAP(A4, A10);
    SWAP(A5, A11);
    SWAP(A5, A10);
    SWAP(A4, A9);
    SWAP(A5, A9);
    SWAP(A3, A6);
    SWAP(A4, A7);
    SWAP(A5, A8);
    SWAP(A5, A7);
    SWAP(A4, A6);
    SWAP(A5, A6);
    SWAP(A13, A14);
    SWAP(A12, A14);
    SWAP(A12, A13);
    SWAP(A16, A17);
    SWAP(A15, A17);
    SWAP(A15, A16);
    SWAP(A12, A15);
    SWAP(A13, A16);
    SWAP(A14, A17);
    SWAP(A14, A16);
    SWAP(A13, A15);
    SWAP(A14, A15);
    SWAP(A19, A20);
    SWAP(A18, A20);
    SWAP(A18, A19);
    SWAP(A21, A22);
    SWAP(A23, A24);
    SWAP(A21, A23);
    SWAP(A22, A24);
    SWAP(A22, A23);
    SWAP(A18, A22);
    SWAP(A18, A21);
    SWAP(A19, A23);
    SWAP(A20, A24);
    SWAP(A20, A23);
    SWAP(A19, A21);
    SWAP(A20, A22);
    SWAP(A20, A21);
    SWAP(A12, A19);
    SWAP(A12, A18);
    SWAP(A13, A20);
    SWAP(A14, A21);
    SWAP(A14, A20);
    SWAP(A13, A18);
    SWAP(A14, A19);
    SWAP(A14, A18);
    SWAP(A15, A22);
    SWAP(A16, A23);
    SWAP(A17, A24);
    SWAP(A17, A23);
    SWAP(A16, A22);
    SWAP(A17, A22);
    SWAP(A15, A19);
    SWAP(A15, A18);
    SWAP(A16, A20);
    SWAP(A17, A21);
    SWAP(A17, A20);
    SWAP(A16, A18);
    SWAP(A17, A19);
    SWAP(A17, A18);
    SWAP(A0, A13);
    SWAP(A0, A12);
    SWAP(A1, A14);
    SWAP(A2, A15);
    SWAP(A2, A14);
    SWAP(A1, A12);
    SWAP(A2, A13);
    SWAP(A2, A12);
    SWAP(A3, A16);
    SWAP(A4, A17);
    SWAP(A5, A18);
    SWAP(A5, A17);
    SWAP(A4, A16);
    SWAP(A5, A16);
    SWAP(A3, A13);
    SWAP(A3, A12);
    SWAP(A4, A14);
    SWAP(A5, A15);
    SWAP(A5, A14);
    SWAP(A4, A12);
    SWAP(A5, A13);
    SWAP(A5, A12);
    SWAP(A6, A19);
    SWAP(A7, A20);
    SWAP(A8, A21);
    SWAP(A8, A20);
    SWAP(A7, A19);
    SWAP(A8, A19);
    SWAP(A9, A22);
    SWAP(A10, A23);
    SWAP(A11, A24);
    SWAP(A11, A23);
    SWAP(A10, A22);
    SWAP(A11, A22);
    SWAP(A9, A19);
    SWAP(A10, A20);
    SWAP(A11, A21);
    SWAP(A11, A20);
    SWAP(A10, A19);
    SWAP(A11, A19);
    SWAP(A6, A13);
    SWAP(A6, A12);
    SWAP(A7, A14);
    SWAP(A8, A15);
    SWAP(A8, A14);
    SWAP(A7, A12);
    SWAP(A8, A13);
    SWAP(A8, A12);
    SWAP(A9, A16);
    SWAP(A10, A17);
    SWAP(A11, A18);
    SWAP(A11, A17);
    SWAP(A10, A16);
    SWAP(A11, A16);
    SWAP(A9, A13);
    SWAP(A9, A12);
    SWAP(A10, A14);
    SWAP(A11, A15);
    SWAP(A11, A14);
    SWAP(A10, A12);
    SWAP(A11, A13);
    SWAP(A11, A12);
}

template <typename ValueType>
static inline
void sort26(ValueType& A0, ValueType& A1, ValueType& A2, ValueType& A3, ValueType& A4, ValueType& A5, ValueType& A6, ValueType& A7, ValueType& A8, ValueType& A9, ValueType& A10, ValueType& A11, ValueType& A12, ValueType& A13, ValueType& A14, ValueType& A15, ValueType& A16, ValueType& A17, ValueType& A18, ValueType& A19, ValueType& A20, ValueType& A21, ValueType& A22, ValueType& A23, ValueType& A24, ValueType& A25)
{
    SWAP(A1, A2);
    SWAP(A0, A2);
    SWAP(A0, A1);
    SWAP(A4, A5);
    SWAP(A3, A5);
    SWAP(A3, A4);
    SWAP(A0, A3);
    SWAP(A1, A4);
    SWAP(A2, A5);
    SWAP(A2, A4);
    SWAP(A1, A3);
    SWAP(A2, A3);
    SWAP(A7, A8);
    SWAP(A6, A8);
    SWAP(A6, A7);
    SWAP(A9, A10);
    SWAP(A11, A12);
    SWAP(A9, A11);
    SWAP(A10, A12);
    SWAP(A10, A11);
    SWAP(A6, A10);
    SWAP(A6, A9);
    SWAP(A7, A11);
    SWAP(A8, A12);
    SWAP(A8, A11);
    SWAP(A7, A9);
    SWAP(A8, A10);
    SWAP(A8, A9);
    SWAP(A0, A7);
    SWAP(A0, A6);
    SWAP(A1, A8);
    SWAP(A2, A9);
    SWAP(A2, A8);
    SWAP(A1, A6);
    SWAP(A2, A7);
    SWAP(A2, A6);
    SWAP(A3, A10);
    SWAP(A4, A11);
    SWAP(A5, A12);
    SWAP(A5, A11);
    SWAP(A4, A10);
    SWAP(A5, A10);
    SWAP(A3, A7);
    SWAP(A3, A6);
    SWAP(A4, A8);
    SWAP(A5, A9);
    SWAP(A5, A8);
    SWAP(A4, A6);
    SWAP(A5, A7);
    SWAP(A5, A6);
    SWAP(A14, A15);
    SWAP(A13, A15);
    SWAP(A13, A14);
    SWAP(A17, A18);
    SWAP(A16, A18);
    SWAP(A16, A17);
    SWAP(A13, A16);
    SWAP(A14, A17);
    SWAP(A15, A18);
    SWAP(A15, A17);
    SWAP(A14, A16);
    SWAP(A15, A16);
    SWAP(A20, A21);
    SWAP(A19, A21);
    SWAP(A19, A20);
    SWAP(A22, A23);
    SWAP(A24, A25);
    SWAP(A22, A24);
    SWAP(A23, A25);
    SWAP(A23, A24);
    SWAP(A19, A23);
    SWAP(A19, A22);
    SWAP(A20, A24);
    SWAP(A21, A25);
    SWAP(A21, A24);
    SWAP(A20, A22);
    SWAP(A21, A23);
    SWAP(A21, A22);
    SWAP(A13, A20);
    SWAP(A13, A19);
    SWAP(A14, A21);
    SWAP(A15, A22);
    SWAP(A15, A21);
    SWAP(A14, A19);
    SWAP(A15, A20);
    SWAP(A15, A19);
    SWAP(A16, A23);
    SWAP(A17, A24);
    SWAP(A18, A25);
    SWAP(A18, A24);
    SWAP(A17, A23);
    SWAP(A18, A23);
    SWAP(A16, A20);
    SWAP(A16, A19);
    SWAP(A17, A21);
    SWAP(A18, A22);
    SWAP(A18, A21);
    SWAP(A17, A19);
    SWAP(A18, A20);
    SWAP(A18, A19);
    SWAP(A0, A13);
    SWAP(A1, A14);
    SWAP(A2, A15);
    SWAP(A2, A14);
    SWAP(A1, A13);
    SWAP(A2, A13);
    SWAP(A3, A16);
    SWAP(A4, A17);
    SWAP(A5, A18);
    SWAP(A5, A17);
    SWAP(A4, A16);
    SWAP(A5, A16);
    SWAP(A3, A13);
    SWAP(A4, A14);
    SWAP(A5, A15);
    SWAP(A5, A14);
    SWAP(A4, A13);
    SWAP(A5, A13);
    SWAP(A6, A19);
    SWAP(A7, A20);
    SWAP(A8, A21);
    SWAP(A8, A20);
    SWAP(A7, A19);
    SWAP(A8, A19);
    SWAP(A9, A22);
    SWAP(A10, A23);
    SWAP(A10, A22);
    SWAP(A11, A24);
    SWAP(A12, A25);
    SWAP(A12, A24);
    SWAP(A11, A22);
    SWAP(A12, A23);
    SWAP(A12, A22);
    SWAP(A9, A19);
    SWAP(A10, A20);
    SWAP(A10, A19);
    SWAP(A11, A21);
    SWAP(A12, A21);
    SWAP(A11, A19);
    SWAP(A12, A20);
    SWAP(A12, A19);
    SWAP(A6, A13);
    SWAP(A7, A14);
    SWAP(A8, A15);
    SWAP(A8, A14);
    SWAP(A7, A13);
    SWAP(A8, A13);
    SWAP(A9, A16);
    SWAP(A10, A17);
    SWAP(A10, A16);
    SWAP(A11, A18);
    SWAP(A12, A18);
    SWAP(A11, A16);
    SWAP(A12, A17);
    SWAP(A12, A16);
    SWAP(A9, A13);
    SWAP(A10, A14);
    SWAP(A10, A13);
    SWAP(A11, A15);
    SWAP(A12, A15);
    SWAP(A11, A13);
    SWAP(A12, A14);
    SWAP(A12, A13);
}

template <typename ValueType>
static inline
void sort27(ValueType& A0, ValueType& A1, ValueType& A2, ValueType& A3, ValueType& A4, ValueType& A5, ValueType& A6, ValueType& A7, ValueType& A8, ValueType& A9, ValueType& A10, ValueType& A11, ValueType& A12, ValueType& A13, ValueType& A14, ValueType& A15, ValueType& A16, ValueType& A17, ValueType& A18, ValueType& A19, ValueType& A20, ValueType& A21, ValueType& A22, ValueType& A23, ValueType& A24, ValueType& A25, ValueType& A26)
{
    SWAP(A1, A2);
    SWAP(A0, A2);
    SWAP(A0, A1);
    SWAP(A4, A5);
    SWAP(A3, A5);
    SWAP(A3, A4);
    SWAP(A0, A3);
    SWAP(A1, A4);
    SWAP(A2, A5);
    SWAP(A2, A4);
    SWAP(A1, A3);
    SWAP(A2, A3);
    SWAP(A7, A8);
    SWAP(A6, A8);
    SWAP(A6, A7);
    SWAP(A9, A10);
    SWAP(A11, A12);
    SWAP(A9, A11);
    SWAP(A10, A12);
    SWAP(A10, A11);
    SWAP(A6, A10);
    SWAP(A6, A9);
    SWAP(A7, A11);
    SWAP(A8, A12);
    SWAP(A8, A11);
    SWAP(A7, A9);
    SWAP(A8, A10);
    SWAP(A8, A9);
    SWAP(A0, A7);
    SWAP(A0, A6);
    SWAP(A1, A8);
    SWAP(A2, A9);
    SWAP(A2, A8);
    SWAP(A1, A6);
    SWAP(A2, A7);
    SWAP(A2, A6);
    SWAP(A3, A10);
    SWAP(A4, A11);
    SWAP(A5, A12);
    SWAP(A5, A11);
    SWAP(A4, A10);
    SWAP(A5, A10);
    SWAP(A3, A7);
    SWAP(A3, A6);
    SWAP(A4, A8);
    SWAP(A5, A9);
    SWAP(A5, A8);
    SWAP(A4, A6);
    SWAP(A5, A7);
    SWAP(A5, A6);
    SWAP(A14, A15);
    SWAP(A13, A15);
    SWAP(A13, A14);
    SWAP(A16, A17);
    SWAP(A18, A19);
    SWAP(A16, A18);
    SWAP(A17, A19);
    SWAP(A17, A18);
    SWAP(A13, A17);
    SWAP(A13, A16);
    SWAP(A14, A18);
    SWAP(A15, A19);
    SWAP(A15, A18);
    SWAP(A14, A16);
    SWAP(A15, A17);
    SWAP(A15, A16);
    SWAP(A21, A22);
    SWAP(A20, A22);
    SWAP(A20, A21);
    SWAP(A23, A24);
    SWAP(A25, A26);
    SWAP(A23, A25);
    SWAP(A24, A26);
    SWAP(A24, A25);
    SWAP(A20, A24);
    SWAP(A20, A23);
    SWAP(A21, A25);
    SWAP(A22, A26);
    SWAP(A22, A25);
    SWAP(A21, A23);
    SWAP(A22, A24);
    SWAP(A22, A23);
    SWAP(A13, A20);
    SWAP(A14, A21);
    SWAP(A15, A22);
    SWAP(A15, A21);
    SWAP(A14, A20);
    SWAP(A15, A20);
    SWAP(A16, A23);
    SWAP(A17, A24);
    SWAP(A17, A23);
    SWAP(A18, A25);
    SWAP(A19, A26);
    SWAP(A19, A25);
    SWAP(A18, A23);
    SWAP(A19, A24);
    SWAP(A19, A23);
    SWAP(A16, A20);
    SWAP(A17, A21);
    SWAP(A17, A20);
    SWAP(A18, A22);
    SWAP(A19, A22);
    SWAP(A18, A20);
    SWAP(A19, A21);
    SWAP(A19, A20);
    SWAP(A0, A14);
    SWAP(A0, A13);
    SWAP(A1, A15);
    SWAP(A2, A16);
    SWAP(A2, A15);
    SWAP(A1, A13);
    SWAP(A2, A14);
    SWAP(A2, A13);
    SWAP(A3, A17);
    SWAP(A4, A18);
    SWAP(A5, A19);
    SWAP(A5, A18);
    SWAP(A4, A17);
    SWAP(A5, A17);
    SWAP(A3, A14);
    SWAP(A3, A13);
    SWAP(A4, A15);
    SWAP(A5, A16);
    SWAP(A5, A15);
    SWAP(A4, A13);
    SWAP(A5, A14);
    SWAP(A5, A13);
    SWAP(A6, A20);
    SWAP(A7, A21);
    SWAP(A8, A22);
    SWAP(A8, A21);
    SWAP(A7, A20);
    SWAP(A8, A20);
    SWAP(A9, A23);
    SWAP(A10, A24);
    SWAP(A10, A23);
    SWAP(A11, A25);
    SWAP(A12, A26);
    SWAP(A12, A25);
    SWAP(A11, A23);
    SWAP(A12, A24);
    SWAP(A12, A23);
    SWAP(A9, A20);
    SWAP(A10, A21);
    SWAP(A10, A20);
    SWAP(A11, A22);
    SWAP(A12, A22);
    SWAP(A11, A20);
    SWAP(A12, A21);
    SWAP(A12, A20);
    SWAP(A6, A13);
    SWAP(A7, A14);
    SWAP(A8, A15);
    SWAP(A8, A14);
    SWAP(A7, A13);
    SWAP(A8, A13);
    SWAP(A9, A16);
    SWAP(A10, A17);
    SWAP(A10, A16);
    SWAP(A11, A18);
    SWAP(A12, A19);
    SWAP(A12, A18);
    SWAP(A11, A16);
    SWAP(A12, A17);
    SWAP(A12, A16);
    SWAP(A9, A13);
    SWAP(A10, A14);
    SWAP(A10, A13);
    SWAP(A11, A15);
    SWAP(A12, A15);
    SWAP(A11, A13);
    SWAP(A12, A14);
    SWAP(A12, A13);
}

template <typename ValueType>
static inline
void sort28(ValueType& A0, ValueType& A1, ValueType& A2, ValueType& A3, ValueType& A4, ValueType& A5, ValueType& A6, ValueType& A7, ValueType& A8, ValueType& A9, ValueType& A10, ValueType& A11, ValueType& A12, ValueType& A13, ValueType& A14, ValueType& A15, ValueType& A16, ValueType& A17, ValueType& A18, ValueType& A19, ValueType& A20, ValueType& A21, ValueType& A22, ValueType& A23, ValueType& A24, ValueType& A25, ValueType& A26, ValueType& A27)
{
    SWAP(A1, A2);
    SWAP(A0, A2);
    SWAP(A0, A1);
    SWAP(A3, A4);
    SWAP(A5, A6);
    SWAP(A3, A5);
    SWAP(A4, A6);
    SWAP(A4, A5);
    SWAP(A0, A4);
    SWAP(A0, A3);
    SWAP(A1, A5);
    SWAP(A2, A6);
    SWAP(A2, A5);
    SWAP(A1, A3);
    SWAP(A2, A4);
    SWAP(A2, A3);
    SWAP(A8, A9);
    SWAP(A7, A9);
    SWAP(A7, A8);
    SWAP(A10, A11);
    SWAP(A12, A13);
    SWAP(A10, A12);
    SWAP(A11, A13);
    SWAP(A11, A12);
    SWAP(A7, A11);
    SWAP(A7, A10);
    SWAP(A8, A12);
    SWAP(A9, A13);
    SWAP(A9, A12);
    SWAP(A8, A10);
    SWAP(A9, A11);
    SWAP(A9, A10);
    SWAP(A0, A7);
    SWAP(A1, A8);
    SWAP(A2, A9);
    SWAP(A2, A8);
    SWAP(A1, A7);
    SWAP(A2, A7);
    SWAP(A3, A10);
    SWAP(A4, A11);
    SWAP(A4, A10);
    SWAP(A5, A12);
    SWAP(A6, A13);
    SWAP(A6, A12);
    SWAP(A5, A10);
    SWAP(A6, A11);
    SWAP(A6, A10);
    SWAP(A3, A7);
    SWAP(A4, A8);
    SWAP(A4, A7);
    SWAP(A5, A9);
    SWAP(A6, A9);
    SWAP(A5, A7);
    SWAP(A6, A8);
    SWAP(A6, A7);
    SWAP(A15, A16);
    SWAP(A14, A16);
    SWAP(A14, A15);
    SWAP(A17, A18);
    SWAP(A19, A20);
    SWAP(A17, A19);
    SWAP(A18, A20);
    SWAP(A18, A19);
    SWAP(A14, A18);
    SWAP(A14, A17);
    SWAP(A15, A19);
    SWAP(A16, A20);
    SWAP(A16, A19);
    SWAP(A15, A17);
    SWAP(A16, A18);
    SWAP(A16, A17);
    SWAP(A22, A23);
    SWAP(A21, A23);
    SWAP(A21, A22);
    SWAP(A24, A25);
    SWAP(A26, A27);
    SWAP(A24, A26);
    SWAP(A25, A27);
    SWAP(A25, A26);
    SWAP(A21, A25);
    SWAP(A21, A24);
    SWAP(A22, A26);
    SWAP(A23, A27);
    SWAP(A23, A26);
    SWAP(A22, A24);
    SWAP(A23, A25);
    SWAP(A23, A24);
    SWAP(A14, A21);
    SWAP(A15, A22);
    SWAP(A16, A23);
    SWAP(A16, A22);
    SWAP(A15, A21);
    SWAP(A16, A21);
    SWAP(A17, A24);
    SWAP(A18, A25);
    SWAP(A18, A24);
    SWAP(A19, A26);
    SWAP(A20, A27);
    SWAP(A20, A26);
    SWAP(A19, A24);
    SWAP(A20, A25);
    SWAP(A20, A24);
    SWAP(A17, A21);
    SWAP(A18, A22);
    SWAP(A18, A21);
    SWAP(A19, A23);
    SWAP(A20, A23);
    SWAP(A19, A21);
    SWAP(A20, A22);
    SWAP(A20, A21);
    SWAP(A0, A14);
    SWAP(A1, A15);
    SWAP(A2, A16);
    SWAP(A2, A15);
    SWAP(A1, A14);
    SWAP(A2, A14);
    SWAP(A3, A17);
    SWAP(A4, A18);
    SWAP(A4, A17);
    SWAP(A5, A19);
    SWAP(A6, A20);
    SWAP(A6, A19);
    SWAP(A5, A17);
    SWAP(A6, A18);
    SWAP(A6, A17);
    SWAP(A3, A14);
    SWAP(A4, A15);
    SWAP(A4, A14);
    SWAP(A5, A16);
    SWAP(A6, A16);
    SWAP(A5, A14);
    SWAP(A6, A15);
    SWAP(A6, A14);
    SWAP(A7, A21);
    SWAP(A8, A22);
    SWAP(A9, A23);
    SWAP(A9, A22);
    SWAP(A8, A21);
    SWAP(A9, A21);
    SWAP(A10, A24);
    SWAP(A11, A25);
    SWAP(A11, A24);
    SWAP(A12, A26);
    SWAP(A13, A27);
    SWAP(A13, A26);
    SWAP(A12, A24);
    SWAP(A13, A25);
    SWAP(A13, A24);
    SWAP(A10, A21);
    SWAP(A11, A22);
    SWAP(A11, A21);
    SWAP(A12, A23);
    SWAP(A13, A23);
    SWAP(A12, A21);
    SWAP(A13, A22);
    SWAP(A13, A21);
    SWAP(A7, A14);
    SWAP(A8, A15);
    SWAP(A9, A16);
    SWAP(A9, A15);
    SWAP(A8, A14);
    SWAP(A9, A14);
    SWAP(A10, A17);
    SWAP(A11, A18);
    SWAP(A11, A17);
    SWAP(A12, A19);
    SWAP(A13, A20);
    SWAP(A13, A19);
    SWAP(A12, A17);
    SWAP(A13, A18);
    SWAP(A13, A17);
    SWAP(A10, A14);
    SWAP(A11, A15);
    SWAP(A11, A14);
    SWAP(A12, A16);
    SWAP(A13, A16);
    SWAP(A12, A14);
    SWAP(A13, A15);
    SWAP(A13, A14);
}

template <typename ValueType>
static inline
void sort29(ValueType& A0, ValueType& A1, ValueType& A2, ValueType& A3, ValueType& A4, ValueType& A5, ValueType& A6, ValueType& A7, ValueType& A8, ValueType& A9, ValueType& A10, ValueType& A11, ValueType& A12, ValueType& A13, ValueType& A14, ValueType& A15, ValueType& A16, ValueType& A17, ValueType& A18, ValueType& A19, ValueType& A20, ValueType& A21, ValueType& A22, ValueType& A23, ValueType& A24, ValueType& A25, ValueType& A26, ValueType& A27, ValueType& A28)
{
    SWAP(A1, A2);
    SWAP(A0, A2);
    SWAP(A0, A1);
    SWAP(A3, A4);
    SWAP(A5, A6);
    SWAP(A3, A5);
    SWAP(A4, A6);
    SWAP(A4, A5);
    SWAP(A0, A4);
    SWAP(A0, A3);
    SWAP(A1, A5);
    SWAP(A2, A6);
    SWAP(A2, A5);
    SWAP(A1, A3);
    SWAP(A2, A4);
    SWAP(A2, A3);
    SWAP(A8, A9);
    SWAP(A7, A9);
    SWAP(A7, A8);
    SWAP(A10, A11);
    SWAP(A12, A13);
    SWAP(A10, A12);
    SWAP(A11, A13);
    SWAP(A11, A12);
    SWAP(A7, A11);
    SWAP(A7, A10);
    SWAP(A8, A12);
    SWAP(A9, A13);
    SWAP(A9, A12);
    SWAP(A8, A10);
    SWAP(A9, A11);
    SWAP(A9, A10);
    SWAP(A0, A7);
    SWAP(A1, A8);
    SWAP(A2, A9);
    SWAP(A2, A8);
    SWAP(A1, A7);
    SWAP(A2, A7);
    SWAP(A3, A10);
    SWAP(A4, A11);
    SWAP(A4, A10);
    SWAP(A5, A12);
    SWAP(A6, A13);
    SWAP(A6, A12);
    SWAP(A5, A10);
    SWAP(A6, A11);
    SWAP(A6, A10);
    SWAP(A3, A7);
    SWAP(A4, A8);
    SWAP(A4, A7);
    SWAP(A5, A9);
    SWAP(A6, A9);
    SWAP(A5, A7);
    SWAP(A6, A8);
    SWAP(A6, A7);
    SWAP(A15, A16);
    SWAP(A14, A16);
    SWAP(A14, A15);
    SWAP(A17, A18);
    SWAP(A19, A20);
    SWAP(A17, A19);
    SWAP(A18, A20);
    SWAP(A18, A19);
    SWAP(A14, A18);
    SWAP(A14, A17);
    SWAP(A15, A19);
    SWAP(A16, A20);
    SWAP(A16, A19);
    SWAP(A15, A17);
    SWAP(A16, A18);
    SWAP(A16, A17);
    SWAP(A21, A22);
    SWAP(A23, A24);
    SWAP(A21, A23);
    SWAP(A22, A24);
    SWAP(A22, A23);
    SWAP(A25, A26);
    SWAP(A27, A28);
    SWAP(A25, A27);
    SWAP(A26, A28);
    SWAP(A26, A27);
    SWAP(A21, A25);
    SWAP(A22, A26);
    SWAP(A22, A25);
    SWAP(A23, A27);
    SWAP(A24, A28);
    SWAP(A24, A27);
    SWAP(A23, A25);
    SWAP(A24, A26);
    SWAP(A24, A25);
    SWAP(A14, A22);
    SWAP(A14, A21);
    SWAP(A15, A23);
    SWAP(A16, A24);
    SWAP(A16, A23);
    SWAP(A15, A21);
    SWAP(A16, A22);
    SWAP(A16, A21);
    SWAP(A17, A25);
    SWAP(A18, A26);
    SWAP(A18, A25);
    SWAP(A19, A27);
    SWAP(A20, A28);
    SWAP(A20, A27);
    SWAP(A19, A25);
    SWAP(A20, A26);
    SWAP(A20, A25);
    SWAP(A17, A21);
    SWAP(A18, A22);
    SWAP(A18, A21);
    SWAP(A19, A23);
    SWAP(A20, A24);
    SWAP(A20, A23);
    SWAP(A19, A21);
    SWAP(A20, A22);
    SWAP(A20, A21);
    SWAP(A0, A15);
    SWAP(A0, A14);
    SWAP(A1, A16);
    SWAP(A2, A17);
    SWAP(A2, A16);
    SWAP(A1, A14);
    SWAP(A2, A15);
    SWAP(A2, A14);
    SWAP(A3, A18);
    SWAP(A4, A19);
    SWAP(A4, A18);
    SWAP(A5, A20);
    SWAP(A6, A21);
    SWAP(A6, A20);
    SWAP(A5, A18);
    SWAP(A6, A19);
    SWAP(A6, A18);
    SWAP(A3, A14);
    SWAP(A4, A15);
    SWAP(A4, A14);
    SWAP(A5, A16);
    SWAP(A6, A17);
    SWAP(A6, A16);
    SWAP(A5, A14);
    SWAP(A6, A15);
    SWAP(A6, A14);
    SWAP(A7, A22);
    SWAP(A8, A23);
    SWAP(A9, A24);
    SWAP(A9, A23);
    SWAP(A8, A22);
    SWAP(A9, A22);
    SWAP(A10, A25);
    SWAP(A11, A26);
    SWAP(A11, A25);
    SWAP(A12, A27);
    SWAP(A13, A28);
    SWAP(A13, A27);
    SWAP(A12, A25);
    SWAP(A13, A26);
    SWAP(A13, A25);
    SWAP(A10, A22);
    SWAP(A11, A23);
    SWAP(A11, A22);
    SWAP(A12, A24);
    SWAP(A13, A24);
    SWAP(A12, A22);
    SWAP(A13, A23);
    SWAP(A13, A22);
    SWAP(A7, A15);
    SWAP(A7, A14);
    SWAP(A8, A16);
    SWAP(A9, A17);
    SWAP(A9, A16);
    SWAP(A8, A14);
    SWAP(A9, A15);
    SWAP(A9, A14);
    SWAP(A10, A18);
    SWAP(A11, A19);
    SWAP(A11, A18);
    SWAP(A12, A20);
    SWAP(A13, A21);
    SWAP(A13, A20);
    SWAP(A12, A18);
    SWAP(A13, A19);
    SWAP(A13, A18);
    SWAP(A10, A14);
    SWAP(A11, A15);
    SWAP(A11, A14);
    SWAP(A12, A16);
    SWAP(A13, A17);
    SWAP(A13, A16);
    SWAP(A12, A14);
    SWAP(A13, A15);
    SWAP(A13, A14);
}

template <typename ValueType>
static inline
void sort30(ValueType& A0, ValueType& A1, ValueType& A2, ValueType& A3, ValueType& A4, ValueType& A5, ValueType& A6, ValueType& A7, ValueType& A8, ValueType& A9, ValueType& A10, ValueType& A11, ValueType& A12, ValueType& A13, ValueType& A14, ValueType& A15, ValueType& A16, ValueType& A17, ValueType& A18, ValueType& A19, ValueType& A20, ValueType& A21, ValueType& A22, ValueType& A23, ValueType& A24, ValueType& A25, ValueType& A26, ValueType& A27, ValueType& A28, ValueType& A29)
{
    SWAP(A1, A2);
    SWAP(A0, A2);
    SWAP(A0, A1);
    SWAP(A3, A4);
    SWAP(A5, A6);
    SWAP(A3, A5);
    SWAP(A4, A6);
    SWAP(A4, A5);
    SWAP(A0, A4);
    SWAP(A0, A3);
    SWAP(A1, A5);
    SWAP(A2, A6);
    SWAP(A2, A5);
    SWAP(A1, A3);
    SWAP(A2, A4);
    SWAP(A2, A3);
    SWAP(A7, A8);
    SWAP(A9, A10);
    SWAP(A7, A9);
    SWAP(A8, A10);
    SWAP(A8, A9);
    SWAP(A11, A12);
    SWAP(A13, A14);
    SWAP(A11, A13);
    SWAP(A12, A14);
    SWAP(A12, A13);
    SWAP(A7, A11);
    SWAP(A8, A12);
    SWAP(A8, A11);
    SWAP(A9, A13);
    SWAP(A10, A14);
    SWAP(A10, A13);
    SWAP(A9, A11);
    SWAP(A10, A12);
    SWAP(A10, A11);
    SWAP(A0, A8);
    SWAP(A0, A7);
    SWAP(A1, A9);
    SWAP(A2, A10);
    SWAP(A2, A9);
    SWAP(A1, A7);
    SWAP(A2, A8);
    SWAP(A2, A7);
    SWAP(A3, A11);
    SWAP(A4, A12);
    SWAP(A4, A11);
    SWAP(A5, A13);
    SWAP(A6, A14);
    SWAP(A6, A13);
    SWAP(A5, A11);
    SWAP(A6, A12);
    SWAP(A6, A11);
    SWAP(A3, A7);
    SWAP(A4, A8);
    SWAP(A4, A7);
    SWAP(A5, A9);
    SWAP(A6, A10);
    SWAP(A6, A9);
    SWAP(A5, A7);
    SWAP(A6, A8);
    SWAP(A6, A7);
    SWAP(A16, A17);
    SWAP(A15, A17);
    SWAP(A15, A16);
    SWAP(A18, A19);
    SWAP(A20, A21);
    SWAP(A18, A20);
    SWAP(A19, A21);
    SWAP(A19, A20);
    SWAP(A15, A19);
    SWAP(A15, A18);
    SWAP(A16, A20);
    SWAP(A17, A21);
    SWAP(A17, A20);
    SWAP(A16, A18);
    SWAP(A17, A19);
    SWAP(A17, A18);
    SWAP(A22, A23);
    SWAP(A24, A25);
    SWAP(A22, A24);
    SWAP(A23, A25);
    SWAP(A23, A24);
    SWAP(A26, A27);
    SWAP(A28, A29);
    SWAP(A26, A28);
    SWAP(A27, A29);
    SWAP(A27, A28);
    SWAP(A22, A26);
    SWAP(A23, A27);
    SWAP(A23, A26);
    SWAP(A24, A28);
    SWAP(A25, A29);
    SWAP(A25, A28);
    SWAP(A24, A26);
    SWAP(A25, A27);
    SWAP(A25, A26);
    SWAP(A15, A23);
    SWAP(A15, A22);
    SWAP(A16, A24);
    SWAP(A17, A25);
    SWAP(A17, A24);
    SWAP(A16, A22);
    SWAP(A17, A23);
    SWAP(A17, A22);
    SWAP(A18, A26);
    SWAP(A19, A27);
    SWAP(A19, A26);
    SWAP(A20, A28);
    SWAP(A21, A29);
    SWAP(A21, A28);
    SWAP(A20, A26);
    SWAP(A21, A27);
    SWAP(A21, A26);
    SWAP(A18, A22);
    SWAP(A19, A23);
    SWAP(A19, A22);
    SWAP(A20, A24);
    SWAP(A21, A25);
    SWAP(A21, A24);
    SWAP(A20, A22);
    SWAP(A21, A23);
    SWAP(A21, A22);
    SWAP(A0, A15);
    SWAP(A1, A16);
    SWAP(A2, A17);
    SWAP(A2, A16);
    SWAP(A1, A15);
    SWAP(A2, A15);
    SWAP(A3, A18);
    SWAP(A4, A19);
    SWAP(A4, A18);
    SWAP(A5, A20);
    SWAP(A6, A21);
    SWAP(A6, A20);
    SWAP(A5, A18);
    SWAP(A6, A19);
    SWAP(A6, A18);
    SWAP(A3, A15);
    SWAP(A4, A16);
    SWAP(A4, A15);
    SWAP(A5, A17);
    SWAP(A6, A17);
    SWAP(A5, A15);
    SWAP(A6, A16);
    SWAP(A6, A15);
    SWAP(A7, A22);
    SWAP(A8, A23);
    SWAP(A8, A22);
    SWAP(A9, A24);
    SWAP(A10, A25);
    SWAP(A10, A24);
    SWAP(A9, A22);
    SWAP(A10, A23);
    SWAP(A10, A22);
    SWAP(A11, A26);
    SWAP(A12, A27);
    SWAP(A12, A26);
    SWAP(A13, A28);
    SWAP(A14, A29);
    SWAP(A14, A28);
    SWAP(A13, A26);
    SWAP(A14, A27);
    SWAP(A14, A26);
    SWAP(A11, A22);
    SWAP(A12, A23);
    SWAP(A12, A22);
    SWAP(A13, A24);
    SWAP(A14, A25);
    SWAP(A14, A24);
    SWAP(A13, A22);
    SWAP(A14, A23);
    SWAP(A14, A22);
    SWAP(A7, A15);
    SWAP(A8, A16);
    SWAP(A8, A15);
    SWAP(A9, A17);
    SWAP(A10, A18);
    SWAP(A10, A17);
    SWAP(A9, A15);
    SWAP(A10, A16);
    SWAP(A10, A15);
    SWAP(A11, A19);
    SWAP(A12, A20);
    SWAP(A12, A19);
    SWAP(A13, A21);
    SWAP(A14, A21);
    SWAP(A13, A19);
    SWAP(A14, A20);
    SWAP(A14, A19);
    SWAP(A11, A15);
    SWAP(A12, A16);
    SWAP(A12, A15);
    SWAP(A13, A17);
    SWAP(A14, A18);
    SWAP(A14, A17);
    SWAP(A13, A15);
    SWAP(A14, A16);
    SWAP(A14, A15);
}

template <typename ValueType>
static inline
void sort31(ValueType& A0, ValueType& A1, ValueType& A2, ValueType& A3, ValueType& A4, ValueType& A5, ValueType& A6, ValueType& A7, ValueType& A8, ValueType& A9, ValueType& A10, ValueType& A11, ValueType& A12, ValueType& A13, ValueType& A14, ValueType& A15, ValueType& A16, ValueType& A17, ValueType& A18, ValueType& A19, ValueType& A20, ValueType& A21, ValueType& A22, ValueType& A23, ValueType& A24, ValueType& A25, ValueType& A26, ValueType& A27, ValueType& A28, ValueType& A29, ValueType& A30)
{
    SWAP(A1, A2);
    SWAP(A0, A2);
    SWAP(A0, A1);
    SWAP(A3, A4);
    SWAP(A5, A6);
    SWAP(A3, A5);
    SWAP(A4, A6);
    SWAP(A4, A5);
    SWAP(A0, A4);
    SWAP(A0, A3);
    SWAP(A1, A5);
    SWAP(A2, A6);
    SWAP(A2, A5);
    SWAP(A1, A3);
    SWAP(A2, A4);
    SWAP(A2, A3);
    SWAP(A7, A8);
    SWAP(A9, A10);
    SWAP(A7, A9);
    SWAP(A8, A10);
    SWAP(A8, A9);
    SWAP(A11, A12);
    SWAP(A13, A14);
    SWAP(A11, A13);
    SWAP(A12, A14);
    SWAP(A12, A13);
    SWAP(A7, A11);
    SWAP(A8, A12);
    SWAP(A8, A11);
    SWAP(A9, A13);
    SWAP(A10, A14);
    SWAP(A10, A13);
    SWAP(A9, A11);
    SWAP(A10, A12);
    SWAP(A10, A11);
    SWAP(A0, A8);
    SWAP(A0, A7);
    SWAP(A1, A9);
    SWAP(A2, A10);
    SWAP(A2, A9);
    SWAP(A1, A7);
    SWAP(A2, A8);
    SWAP(A2, A7);
    SWAP(A3, A11);
    SWAP(A4, A12);
    SWAP(A4, A11);
    SWAP(A5, A13);
    SWAP(A6, A14);
    SWAP(A6, A13);
    SWAP(A5, A11);
    SWAP(A6, A12);
    SWAP(A6, A11);
    SWAP(A3, A7);
    SWAP(A4, A8);
    SWAP(A4, A7);
    SWAP(A5, A9);
    SWAP(A6, A10);
    SWAP(A6, A9);
    SWAP(A5, A7);
    SWAP(A6, A8);
    SWAP(A6, A7);
    SWAP(A15, A16);
    SWAP(A17, A18);
    SWAP(A15, A17);
    SWAP(A16, A18);
    SWAP(A16, A17);
    SWAP(A19, A20);
    SWAP(A21, A22);
    SWAP(A19, A21);
    SWAP(A20, A22);
    SWAP(A20, A21);
    SWAP(A15, A19);
    SWAP(A16, A20);
    SWAP(A16, A19);
    SWAP(A17, A21);
    SWAP(A18, A22);
    SWAP(A18, A21);
    SWAP(A17, A19);
    SWAP(A18, A20);
    SWAP(A18, A19);
    SWAP(A23, A24);
    SWAP(A25, A26);
    SWAP(A23, A25);
    SWAP(A24, A26);
    SWAP(A24, A25);
    SWAP(A27, A28);
    SWAP(A29, A30);
    SWAP(A27, A29);
    SWAP(A28, A30);
    SWAP(A28, A29);
    SWAP(A23, A27);
    SWAP(A24, A28);
    SWAP(A24, A27);
    SWAP(A25, A29);
    SWAP(A26, A30);
    SWAP(A26, A29);
    SWAP(A25, A27);
    SWAP(A26, A28);
    SWAP(A26, A27);
    SWAP(A15, A23);
    SWAP(A16, A24);
    SWAP(A16, A23);
    SWAP(A17, A25);
    SWAP(A18, A26);
    SWAP(A18, A25);
    SWAP(A17, A23);
    SWAP(A18, A24);
    SWAP(A18, A23);
    SWAP(A19, A27);
    SWAP(A20, A28);
    SWAP(A20, A27);
    SWAP(A21, A29);
    SWAP(A22, A30);
    SWAP(A22, A29);
    SWAP(A21, A27);
    SWAP(A22, A28);
    SWAP(A22, A27);
    SWAP(A19, A23);
    SWAP(A20, A24);
    SWAP(A20, A23);
    SWAP(A21, A25);
    SWAP(A22, A26);
    SWAP(A22, A25);
    SWAP(A21, A23);
    SWAP(A22, A24);
    SWAP(A22, A23);
    SWAP(A0, A16);
    SWAP(A0, A15);
    SWAP(A1, A17);
    SWAP(A2, A18);
    SWAP(A2, A17);
    SWAP(A1, A15);
    SWAP(A2, A16);
    SWAP(A2, A15);
    SWAP(A3, A19);
    SWAP(A4, A20);
    SWAP(A4, A19);
    SWAP(A5, A21);
    SWAP(A6, A22);
    SWAP(A6, A21);
    SWAP(A5, A19);
    SWAP(A6, A20);
    SWAP(A6, A19);
    SWAP(A3, A15);
    SWAP(A4, A16);
    SWAP(A4, A15);
    SWAP(A5, A17);
    SWAP(A6, A18);
    SWAP(A6, A17);
    SWAP(A5, A15);
    SWAP(A6, A16);
    SWAP(A6, A15);
    SWAP(A7, A23);
    SWAP(A8, A24);
    SWAP(A8, A23);
    SWAP(A9, A25);
    SWAP(A10, A26);
    SWAP(A10, A25);
    SWAP(A9, A23);
    SWAP(A10, A24);
    SWAP(A10, A23);
    SWAP(A11, A27);
    SWAP(A12, A28);
    SWAP(A12, A27);
    SWAP(A13, A29);
    SWAP(A14, A30);
    SWAP(A14, A29);
    SWAP(A13, A27);
    SWAP(A14, A28);
    SWAP(A14, A27);
    SWAP(A11, A23);
    SWAP(A12, A24);
    SWAP(A12, A23);
    SWAP(A13, A25);
    SWAP(A14, A26);
    SWAP(A14, A25);
    SWAP(A13, A23);
    SWAP(A14, A24);
    SWAP(A14, A23);
    SWAP(A7, A15);
    SWAP(A8, A16);
    SWAP(A8, A15);
    SWAP(A9, A17);
    SWAP(A10, A18);
    SWAP(A10, A17);
    SWAP(A9, A15);
    SWAP(A10, A16);
    SWAP(A10, A15);
    SWAP(A11, A19);
    SWAP(A12, A20);
    SWAP(A12, A19);
    SWAP(A13, A21);
    SWAP(A14, A22);
    SWAP(A14, A21);
    SWAP(A13, A19);
    SWAP(A14, A20);
    SWAP(A14, A19);
    SWAP(A11, A15);
    SWAP(A12, A16);
    SWAP(A12, A15);
    SWAP(A13, A17);
    SWAP(A14, A18);
    SWAP(A14, A17);
    SWAP(A13, A15);
    SWAP(A14, A16);
    SWAP(A14, A15);
}

template <typename ValueType>
static inline
void sort32(ValueType& A0, ValueType& A1, ValueType& A2, ValueType& A3, ValueType& A4, ValueType& A5, ValueType& A6, ValueType& A7, ValueType& A8, ValueType& A9, ValueType& A10, ValueType& A11, ValueType& A12, ValueType& A13, ValueType& A14, ValueType& A15, ValueType& A16, ValueType& A17, ValueType& A18, ValueType& A19, ValueType& A20, ValueType& A21, ValueType& A22, ValueType& A23, ValueType& A24, ValueType& A25, ValueType& A26, ValueType& A27, ValueType& A28, ValueType& A29, ValueType& A30, ValueType& A31)
{
    SWAP(A0, A1);
    SWAP(A2, A3);
    SWAP(A0, A2);
    SWAP(A1, A3);
    SWAP(A1, A2);
    SWAP(A4, A5);
    SWAP(A6, A7);
    SWAP(A4, A6);
    SWAP(A5, A7);
    SWAP(A5, A6);
    SWAP(A0, A4);
    SWAP(A1, A5);
    SWAP(A1, A4);
    SWAP(A2, A6);
    SWAP(A3, A7);
    SWAP(A3, A6);
    SWAP(A2, A4);
    SWAP(A3, A5);
    SWAP(A3, A4);
    SWAP(A8, A9);
    SWAP(A10, A11);
    SWAP(A8, A10);
    SWAP(A9, A11);
    SWAP(A9, A10);
    SWAP(A12, A13);
    SWAP(A14, A15);
    SWAP(A12, A14);
    SWAP(A13, A15);
    SWAP(A13, A14);
    SWAP(A8, A12);
    SWAP(A9, A13);
    SWAP(A9, A12);
    SWAP(A10, A14);
    SWAP(A11, A15);
    SWAP(A11, A14);
    SWAP(A10, A12);
    SWAP(A11, A13);
    SWAP(A11, A12);
    SWAP(A0, A8);
    SWAP(A1, A9);
    SWAP(A1, A8);
    SWAP(A2, A10);
    SWAP(A3, A11);
    SWAP(A3, A10);
    SWAP(A2, A8);
    SWAP(A3, A9);
    SWAP(A3, A8);
    SWAP(A4, A12);
    SWAP(A5, A13);
    SWAP(A5, A12);
    SWAP(A6, A14);
    SWAP(A7, A15);
    SWAP(A7, A14);
    SWAP(A6, A12);
    SWAP(A7, A13);
    SWAP(A7, A12);
    SWAP(A4, A8);
    SWAP(A5, A9);
    SWAP(A5, A8);
    SWAP(A6, A10);
    SWAP(A7, A11);
    SWAP(A7, A10);
    SWAP(A6, A8);
    SWAP(A7, A9);
    SWAP(A7, A8);
    SWAP(A16, A17);
    SWAP(A18, A19);
    SWAP(A16, A18);
    SWAP(A17, A19);
    SWAP(A17, A18);
    SWAP(A20, A21);
    SWAP(A22, A23);
    SWAP(A20, A22);
    SWAP(A21, A23);
    SWAP(A21, A22);
    SWAP(A16, A20);
    SWAP(A17, A21);
    SWAP(A17, A20);
    SWAP(A18, A22);
    SWAP(A19, A23);
    SWAP(A19, A22);
    SWAP(A18, A20);
    SWAP(A19, A21);
    SWAP(A19, A20);
    SWAP(A24, A25);
    SWAP(A26, A27);
    SWAP(A24, A26);
    SWAP(A25, A27);
    SWAP(A25, A26);
    SWAP(A28, A29);
    SWAP(A30, A31);
    SWAP(A28, A30);
    SWAP(A29, A31);
    SWAP(A29, A30);
    SWAP(A24, A28);
    SWAP(A25, A29);
    SWAP(A25, A28);
    SWAP(A26, A30);
    SWAP(A27, A31);
    SWAP(A27, A30);
    SWAP(A26, A28);
    SWAP(A27, A29);
    SWAP(A27, A28);
    SWAP(A16, A24);
    SWAP(A17, A25);
    SWAP(A17, A24);
    SWAP(A18, A26);
    SWAP(A19, A27);
    SWAP(A19, A26);
    SWAP(A18, A24);
    SWAP(A19, A25);
    SWAP(A19, A24);
    SWAP(A20, A28);
    SWAP(A21, A29);
    SWAP(A21, A28);
    SWAP(A22, A30);
    SWAP(A23, A31);
    SWAP(A23, A30);
    SWAP(A22, A28);
    SWAP(A23, A29);
    SWAP(A23, A28);
    SWAP(A20, A24);
    SWAP(A21, A25);
    SWAP(A21, A24);
    SWAP(A22, A26);
    SWAP(A23, A27);
    SWAP(A23, A26);
    SWAP(A22, A24);
    SWAP(A23, A25);
    SWAP(A23, A24);
    SWAP(A0, A16);
    SWAP(A1, A17);
    SWAP(A1, A16);
    SWAP(A2, A18);
    SWAP(A3, A19);
    SWAP(A3, A18);
    SWAP(A2, A16);
    SWAP(A3, A17);
    SWAP(A3, A16);
    SWAP(A4, A20);
    SWAP(A5, A21);
    SWAP(A5, A20);
    SWAP(A6, A22);
    SWAP(A7, A23);
    SWAP(A7, A22);
    SWAP(A6, A20);
    SWAP(A7, A21);
    SWAP(A7, A20);
    SWAP(A4, A16);
    SWAP(A5, A17);
    SWAP(A5, A16);
    SWAP(A6, A18);
    SWAP(A7, A19);
    SWAP(A7, A18);
    SWAP(A6, A16);
    SWAP(A7, A17);
    SWAP(A7, A16);
    SWAP(A8, A24);
    SWAP(A9, A25);
    SWAP(A9, A24);
    SWAP(A10, A26);
    SWAP(A11, A27);
    SWAP(A11, A26);
    SWAP(A10, A24);
    SWAP(A11, A25);
    SWAP(A11, A24);
    SWAP(A12, A28);
    SWAP(A13, A29);
    SWAP(A13, A28);
    SWAP(A14, A30);
    SWAP(A15, A31);
    SWAP(A15, A30);
    SWAP(A14, A28);
    SWAP(A15, A29);
    SWAP(A15, A28);
    SWAP(A12, A24);
    SWAP(A13, A25);
    SWAP(A13, A24);
    SWAP(A14, A26);
    SWAP(A15, A27);
    SWAP(A15, A26);
    SWAP(A14, A24);
    SWAP(A15, A25);
    SWAP(A15, A24);
    SWAP(A8, A16);
    SWAP(A9, A17);
    SWAP(A9, A16);
    SWAP(A10, A18);
    SWAP(A11, A19);
    SWAP(A11, A18);
    SWAP(A10, A16);
    SWAP(A11, A17);
    SWAP(A11, A16);
    SWAP(A12, A20);
    SWAP(A13, A21);
    SWAP(A13, A20);
    SWAP(A14, A22);
    SWAP(A15, A23);
    SWAP(A15, A22);
    SWAP(A14, A20);
    SWAP(A15, A21);
    SWAP(A15, A20);
    SWAP(A12, A16);
    SWAP(A13, A17);
    SWAP(A13, A16);
    SWAP(A14, A18);
    SWAP(A15, A19);
    SWAP(A15, A18);
    SWAP(A14, A16);
    SWAP(A15, A17);
    SWAP(A15, A16);
}



template <typename ValueType> static inline
void sort2(ValueType* A)
{
    SWAP(A[0], A[1]);
}

template <typename ValueType> static inline
void sort3(ValueType* A)
{
    SWAP(A[1], A[2]);
    SWAP(A[0], A[2]);
    SWAP(A[0], A[1]);
}

template <typename ValueType> static inline
void sort4(ValueType* A)
{
    SWAP(A[0], A[1]);
    SWAP(A[2], A[3]);
    SWAP(A[0], A[2]);
    SWAP(A[1], A[3]);
    SWAP(A[1], A[2]);
}

template <typename ValueType> static inline
void sort5(ValueType* A)
{
    SWAP(A[0], A[1]);
    SWAP(A[3], A[4]);
    SWAP(A[2], A[4]);
    SWAP(A[2], A[3]);
    SWAP(A[0], A[3]);
    SWAP(A[0], A[2]);
    SWAP(A[1], A[4]);
    SWAP(A[1], A[3]);
    SWAP(A[1], A[2]);
}

template <typename ValueType> static inline
void sort6(ValueType* A)
{
    SWAP(A[1], A[2]);
    SWAP(A[0], A[2]);
    SWAP(A[0], A[1]);
    SWAP(A[4], A[5]);
    SWAP(A[3], A[5]);
    SWAP(A[3], A[4]);
    SWAP(A[0], A[3]);
    SWAP(A[1], A[4]);
    SWAP(A[2], A[5]);
    SWAP(A[2], A[4]);
    SWAP(A[1], A[3]);
    SWAP(A[2], A[3]);
}

template <typename ValueType> static inline
void sort7(ValueType* A)
{
    SWAP(A[1], A[2]);
    SWAP(A[0], A[2]);
    SWAP(A[0], A[1]);
    SWAP(A[3], A[4]);
    SWAP(A[5], A[6]);
    SWAP(A[3], A[5]);
    SWAP(A[4], A[6]);
    SWAP(A[4], A[5]);
    SWAP(A[0], A[4]);
    SWAP(A[0], A[3]);
    SWAP(A[1], A[5]);
    SWAP(A[2], A[6]);
    SWAP(A[2], A[5]);
    SWAP(A[1], A[3]);
    SWAP(A[2], A[4]);
    SWAP(A[2], A[3]);
}

template <typename ValueType> static inline
void sort8(ValueType* A)
{
    SWAP(A[0], A[1]);
    SWAP(A[2], A[3]);
    SWAP(A[0], A[2]);
    SWAP(A[1], A[3]);
    SWAP(A[1], A[2]);
    SWAP(A[4], A[5]);
    SWAP(A[6], A[7]);
    SWAP(A[4], A[6]);
    SWAP(A[5], A[7]);
    SWAP(A[5], A[6]);
    SWAP(A[0], A[4]);
    SWAP(A[1], A[5]);
    SWAP(A[1], A[4]);
    SWAP(A[2], A[6]);
    SWAP(A[3], A[7]);
    SWAP(A[3], A[6]);
    SWAP(A[2], A[4]);
    SWAP(A[3], A[5]);
    SWAP(A[3], A[4]);
}

template <typename ValueType> static inline
void sort9(ValueType* A)
{
    SWAP(A[0], A[1]);
    SWAP(A[3], A[4]);
    SWAP(A[6], A[7]);
    SWAP(A[1], A[2]);
    SWAP(A[4], A[5]);
    SWAP(A[7], A[8]);
    SWAP(A[0], A[1]);
    SWAP(A[3], A[4]);
    SWAP(A[6], A[7]);
    SWAP(A[0], A[3]);
    SWAP(A[3], A[6]);
    SWAP(A[0], A[3]);
    SWAP(A[1], A[4]);
    SWAP(A[4], A[7]);
    SWAP(A[1], A[4]);
    SWAP(A[2], A[5]);
    SWAP(A[5], A[8]);
    SWAP(A[2], A[5]);
    SWAP(A[1], A[3]);
    SWAP(A[5], A[7]);
    SWAP(A[2], A[6]);
    SWAP(A[4], A[6]);
    SWAP(A[2], A[4]);
    SWAP(A[2], A[3]);
    SWAP(A[5], A[6]);
}

template <typename ValueType> static inline
void sort10(ValueType* A)
{
    SWAP(A[4], A[9]);
    SWAP(A[3], A[8]);
    SWAP(A[2], A[7]);
    SWAP(A[1], A[6]);
    SWAP(A[0], A[5]);
    SWAP(A[1], A[4]);
    SWAP(A[6], A[9]);
    SWAP(A[0], A[3]);
    SWAP(A[5], A[8]);
    SWAP(A[0], A[2]);
    SWAP(A[3], A[6]);
    SWAP(A[7], A[9]);
    SWAP(A[0], A[1]);
    SWAP(A[2], A[4]);
    SWAP(A[5], A[7]);
    SWAP(A[8], A[9]);
    SWAP(A[1], A[2]);
    SWAP(A[4], A[6]);
    SWAP(A[7], A[8]);
    SWAP(A[3], A[5]);
    SWAP(A[2], A[5]);
    SWAP(A[6], A[8]);
    SWAP(A[1], A[3]);
    SWAP(A[4], A[7]);
    SWAP(A[2], A[3]);
    SWAP(A[6], A[7]);
    SWAP(A[3], A[4]);
    SWAP(A[5], A[6]);
    SWAP(A[4], A[5]);
}

template <typename ValueType> static inline
void sort11(ValueType* A)
{
    SWAP(A[0], A[1]);
    SWAP(A[2], A[3]);
    SWAP(A[4], A[5]);
    SWAP(A[6], A[7]);
    SWAP(A[8], A[9]);
    SWAP(A[1], A[3]);
    SWAP(A[5], A[7]);
    SWAP(A[0], A[2]);
    SWAP(A[4], A[6]);
    SWAP(A[8], A[10]);
    SWAP(A[1], A[2]);
    SWAP(A[5], A[6]);
    SWAP(A[9], A[10]);
    SWAP(A[1], A[5]);
    SWAP(A[6], A[10]);
    SWAP(A[5], A[9]);
    SWAP(A[2], A[6]);
    SWAP(A[1], A[5]);
    SWAP(A[6], A[10]);
    SWAP(A[0], A[4]);
    SWAP(A[3], A[7]);
    SWAP(A[4], A[8]);
    SWAP(A[0], A[4]);
    SWAP(A[1], A[4]);
    SWAP(A[7], A[10]);
    SWAP(A[3], A[8]);
    SWAP(A[2], A[3]);
    SWAP(A[8], A[9]);
    SWAP(A[2], A[4]);
    SWAP(A[7], A[9]);
    SWAP(A[3], A[5]);
    SWAP(A[6], A[8]);
    SWAP(A[3], A[4]);
    SWAP(A[5], A[6]);
    SWAP(A[7], A[8]);
}

template <typename ValueType> static inline
void sort12(ValueType* A)
{
    SWAP(A[0], A[1]);
    SWAP(A[2], A[3]);
    SWAP(A[4], A[5]);
    SWAP(A[6], A[7]);
    SWAP(A[8], A[9]);
    SWAP(A[10], A[11]);
    SWAP(A[1], A[3]);
    SWAP(A[5], A[7]);
    SWAP(A[9], A[11]);
    SWAP(A[0], A[2]);
    SWAP(A[4], A[6]);
    SWAP(A[8], A[10]);
    SWAP(A[1], A[2]);
    SWAP(A[5], A[6]);
    SWAP(A[9], A[10]);
    SWAP(A[1], A[5]);
    SWAP(A[6], A[10]);
    SWAP(A[5], A[9]);
    SWAP(A[2], A[6]);
    SWAP(A[1], A[5]);
    SWAP(A[6], A[10]);
    SWAP(A[0], A[4]);
    SWAP(A[7], A[11]);
    SWAP(A[3], A[7]);
    SWAP(A[4], A[8]);
    SWAP(A[0], A[4]);
    SWAP(A[7], A[11]);
    SWAP(A[1], A[4]);
    SWAP(A[7], A[10]);
    SWAP(A[3], A[8]);
    SWAP(A[2], A[3]);
    SWAP(A[8], A[9]);
    SWAP(A[2], A[4]);
    SWAP(A[7], A[9]);
    SWAP(A[3], A[5]);
    SWAP(A[6], A[8]);
    SWAP(A[3], A[4]);
    SWAP(A[5], A[6]);
    SWAP(A[7], A[8]);
}

template <typename ValueType> static inline
void sort13(ValueType* A)
{
    SWAP(A[1], A[7]);
    SWAP(A[9], A[11]);
    SWAP(A[3], A[4]);
    SWAP(A[5], A[8]);
    SWAP(A[0], A[12]);
    SWAP(A[2], A[6]);
    SWAP(A[0], A[1]);
    SWAP(A[2], A[3]);
    SWAP(A[4], A[6]);
    SWAP(A[8], A[11]);
    SWAP(A[7], A[12]);
    SWAP(A[5], A[9]);
    SWAP(A[0], A[2]);
    SWAP(A[3], A[7]);
    SWAP(A[10], A[11]);
    SWAP(A[1], A[4]);
    SWAP(A[6], A[12]);
    SWAP(A[7], A[8]);
    SWAP(A[11], A[12]);
    SWAP(A[4], A[9]);
    SWAP(A[6], A[10]);
    SWAP(A[3], A[4]);
    SWAP(A[5], A[6]);
    SWAP(A[8], A[9]);
    SWAP(A[10], A[11]);
    SWAP(A[1], A[7]);
    SWAP(A[2], A[6]);
    SWAP(A[9], A[11]);
    SWAP(A[1], A[3]);
    SWAP(A[4], A[7]);
    SWAP(A[8], A[10]);
    SWAP(A[0], A[5]);
    SWAP(A[2], A[5]);
    SWAP(A[6], A[8]);
    SWAP(A[9], A[10]);
    SWAP(A[1], A[2]);
    SWAP(A[3], A[5]);
    SWAP(A[7], A[8]);
    SWAP(A[4], A[6]);
    SWAP(A[2], A[3]);
    SWAP(A[4], A[5]);
    SWAP(A[6], A[7]);
    SWAP(A[8], A[9]);
    SWAP(A[3], A[4]);
    SWAP(A[5], A[6]);
}

template <typename ValueType> static inline
void sort14(ValueType* A)
{
    SWAP(A[0], A[1]);
    SWAP(A[2], A[3]);
    SWAP(A[4], A[5]);
    SWAP(A[6], A[7]);
    SWAP(A[8], A[9]);
    SWAP(A[10], A[11]);
    SWAP(A[12], A[13]);
    SWAP(A[0], A[2]);
    SWAP(A[4], A[6]);
    SWAP(A[8], A[10]);
    SWAP(A[1], A[3]);
    SWAP(A[5], A[7]);
    SWAP(A[9], A[11]);
    SWAP(A[0], A[4]);
    SWAP(A[8], A[12]);
    SWAP(A[1], A[5]);
    SWAP(A[9], A[13]);
    SWAP(A[2], A[6]);
    SWAP(A[3], A[7]);
    SWAP(A[0], A[8]);
    SWAP(A[1], A[9]);
    SWAP(A[2], A[10]);
    SWAP(A[3], A[11]);
    SWAP(A[4], A[12]);
    SWAP(A[5], A[13]);
    SWAP(A[5], A[10]);
    SWAP(A[6], A[9]);
    SWAP(A[3], A[12]);
    SWAP(A[7], A[11]);
    SWAP(A[1], A[2]);
    SWAP(A[4], A[8]);
    SWAP(A[1], A[4]);
    SWAP(A[7], A[13]);
    SWAP(A[2], A[8]);
    SWAP(A[2], A[4]);
    SWAP(A[5], A[6]);
    SWAP(A[9], A[10]);
    SWAP(A[11], A[13]);
    SWAP(A[3], A[8]);
    SWAP(A[7], A[12]);
    SWAP(A[6], A[8]);
    SWAP(A[10], A[12]);
    SWAP(A[3], A[5]);
    SWAP(A[7], A[9]);
    SWAP(A[3], A[4]);
    SWAP(A[5], A[6]);
    SWAP(A[7], A[8]);
    SWAP(A[9], A[10]);
    SWAP(A[11], A[12]);
    SWAP(A[6], A[7]);
    SWAP(A[8], A[9]);
}

template <typename ValueType> static inline
void sort15(ValueType* A)
{
    SWAP(A[0], A[1]);
    SWAP(A[2], A[3]);
    SWAP(A[4], A[5]);
    SWAP(A[6], A[7]);
    SWAP(A[8], A[9]);
    SWAP(A[10], A[11]);
    SWAP(A[12], A[13]);
    SWAP(A[0], A[2]);
    SWAP(A[4], A[6]);
    SWAP(A[8], A[10]);
    SWAP(A[12], A[14]);
    SWAP(A[1], A[3]);
    SWAP(A[5], A[7]);
    SWAP(A[9], A[11]);
    SWAP(A[0], A[4]);
    SWAP(A[8], A[12]);
    SWAP(A[1], A[5]);
    SWAP(A[9], A[13]);
    SWAP(A[2], A[6]);
    SWAP(A[10], A[14]);
    SWAP(A[3], A[7]);
    SWAP(A[0], A[8]);
    SWAP(A[1], A[9]);
    SWAP(A[2], A[10]);
    SWAP(A[3], A[11]);
    SWAP(A[4], A[12]);
    SWAP(A[5], A[13]);
    SWAP(A[6], A[14]);
    SWAP(A[5], A[10]);
    SWAP(A[6], A[9]);
    SWAP(A[3], A[12]);
    SWAP(A[13], A[14]);
    SWAP(A[7], A[11]);
    SWAP(A[1], A[2]);
    SWAP(A[4], A[8]);
    SWAP(A[1], A[4]);
    SWAP(A[7], A[13]);
    SWAP(A[2], A[8]);
    SWAP(A[11], A[14]);
    SWAP(A[2], A[4]);
    SWAP(A[5], A[6]);
    SWAP(A[9], A[10]);
    SWAP(A[11], A[13]);
    SWAP(A[3], A[8]);
    SWAP(A[7], A[12]);
    SWAP(A[6], A[8]);
    SWAP(A[10], A[12]);
    SWAP(A[3], A[5]);
    SWAP(A[7], A[9]);
    SWAP(A[3], A[4]);
    SWAP(A[5], A[6]);
    SWAP(A[7], A[8]);
    SWAP(A[9], A[10]);
    SWAP(A[11], A[12]);
    SWAP(A[6], A[7]);
    SWAP(A[8], A[9]);
}

template <typename ValueType>
static inline
void sort16(ValueType* A)
{
    SWAP(A[0], A[1]);
    SWAP(A[2], A[3]);
    SWAP(A[4], A[5]);
    SWAP(A[6], A[7]);
    SWAP(A[8], A[9]);
    SWAP(A[10], A[11]);
    SWAP(A[12], A[13]);
    SWAP(A[14], A[15]);
    SWAP(A[0], A[2]);
    SWAP(A[4], A[6]);
    SWAP(A[8], A[10]);
    SWAP(A[12], A[14]);
    SWAP(A[1], A[3]);
    SWAP(A[5], A[7]);
    SWAP(A[9], A[11]);
    SWAP(A[13], A[15]);
    SWAP(A[0], A[4]);
    SWAP(A[8], A[12]);
    SWAP(A[1], A[5]);
    SWAP(A[9], A[13]);
    SWAP(A[2], A[6]);
    SWAP(A[10], A[14]);
    SWAP(A[3], A[7]);
    SWAP(A[11], A[15]);
    SWAP(A[0], A[8]);
    SWAP(A[1], A[9]);
    SWAP(A[2], A[10]);
    SWAP(A[3], A[11]);
    SWAP(A[4], A[12]);
    SWAP(A[5], A[13]);
    SWAP(A[6], A[14]);
    SWAP(A[7], A[15]);
    SWAP(A[5], A[10]);
    SWAP(A[6], A[9]);
    SWAP(A[3], A[12]);
    SWAP(A[13], A[14]);
    SWAP(A[7], A[11]);
    SWAP(A[1], A[2]);
    SWAP(A[4], A[8]);
    SWAP(A[1], A[4]);
    SWAP(A[7], A[13]);
    SWAP(A[2], A[8]);
    SWAP(A[11], A[14]);
    SWAP(A[2], A[4]);
    SWAP(A[5], A[6]);
    SWAP(A[9], A[10]);
    SWAP(A[11], A[13]);
    SWAP(A[3], A[8]);
    SWAP(A[7], A[12]);
    SWAP(A[6], A[8]);
    SWAP(A[10], A[12]);
    SWAP(A[3], A[5]);
    SWAP(A[7], A[9]);
    SWAP(A[3], A[4]);
    SWAP(A[5], A[6]);
    SWAP(A[7], A[8]);
    SWAP(A[9], A[10]);
    SWAP(A[11], A[12]);
    SWAP(A[6], A[7]);
    SWAP(A[8], A[9]);
}

template <typename ValueType> static inline
void sortN(ValueType* A, size_t n)
{
    switch (n) {
    case 0: return;
    case 1: return;
    case 2: return sort2(A);
    case 3: return sort3(A);
    case 4: return sort4(A);
    case 5: return sort5(A);
    case 6: return sort6(A);
    case 7: return sort7(A);
    case 8: return sort8(A);
    case 9: return sort9(A);
    case 10: return sort10(A);
    case 11: return sort11(A);
    case 12: return sort12(A);
    case 13: return sort13(A);
    case 14: return sort14(A);
    case 15: return sort15(A);
    case 16: return sort16(A);
    }
    abort();
}

} // namespace networks
